from itertools import product
import math
from unittest import mock

import pytest

from localuf import Surface, Repetition
from localuf.decoders import UF
from localuf.type_aliases import Edge
from localuf._determinants import SpaceTimeDeterminant
from localuf._schemes import Forward
from localuf._pairs import LogicalCounter


@pytest.fixture(params=range(3, 11, 2), ids=lambda x: f"d{x}")
def sf_ol(request):
    sf = Surface(
        request.param,
        noise='phenomenological',
        scheme='forward',
    )
    return sf.SCHEME


def test_COMMIT_HEIGHT_attribute(forward_rp: Forward):
    assert forward_rp._COMMIT_HEIGHT == forward_rp._CODE.D


def test_BUFFER_HEIGHT_attribute(forward_rp: Forward):
    assert forward_rp._BUFFER_HEIGHT == forward_rp._CODE.D


def test_DETERMINANT_attribute(forward_rp: Forward):
    assert type(forward_rp._DETERMINANT) is SpaceTimeDeterminant


def test_LOGICAL_COUNTER_attribute(forward3F: Forward):
    assert type(forward3F._LOGICAL_COUNTER) is LogicalCounter


def assert_is_tuple_of_edges(edges: tuple[Edge, ...]):
    assert type(edges) is tuple
    assert type(edges[0]) is tuple
    assert type(edges[0][0]) is tuple
    assert type(edges[0][0][0]) is int


def test_COMMIT_EDGES_attribute(forward_rp: Forward):
    assert_is_tuple_of_edges(forward_rp._COMMIT_EDGES)
    assert set(forward_rp._COMMIT_EDGES).issubset(set(forward_rp._CODE.EDGES))
    assert 2*len(forward_rp._COMMIT_EDGES) == len(forward_rp._CODE.EDGES)


def test_BUFFER_EDGES_attribute(forward_rp: Forward):
    assert_is_tuple_of_edges(forward_rp._BUFFER_EDGES)
    assert set(forward_rp._BUFFER_EDGES).issubset(set(forward_rp._CODE.EDGES))
    assert 2*len(forward_rp._BUFFER_EDGES) == len(forward_rp._CODE.EDGES)
    assert set(forward_rp._COMMIT_EDGES).isdisjoint(set(forward_rp._BUFFER_EDGES))


def test_reset(forward_rp: Forward):
    forward_rp.history = [(set(forward_rp._CODE.EDGES), set(forward_rp._CODE.EDGES), set(forward_rp._CODE.NODES))]
    forward_rp.reset()
    assert not hasattr(forward_rp, 'history')


def test_get_leftover(forward_rp: Forward):
    all_edges = set(forward_rp._CODE.EDGES)
    assert forward_rp._get_leftover(set(), set()) == (set(), set())
    assert forward_rp._get_leftover(set(), all_edges) == (set(forward_rp._COMMIT_EDGES), set())
    assert forward_rp._get_leftover(all_edges, set()) == (set(forward_rp._COMMIT_EDGES), set(forward_rp._BUFFER_EDGES))
    assert forward_rp._get_leftover(all_edges, all_edges) == (set(), set(forward_rp._BUFFER_EDGES))


def test_get_logical_error_repetition(forward_rp: Forward):
    d = forward_rp._CODE.D
    ts = [0, forward_rp._COMMIT_HEIGHT-1]
    paths = [{((j, t), (j+1, t)) for j in range(-1, d-1)} for t in ts]
    assert forward_rp.get_logical_error(set()) == 0
    assert forward_rp.pairs._dc == {}
    assert forward_rp.get_logical_error(paths[0]) == 1
    assert forward_rp.pairs._dc == {}
    assert forward_rp.get_logical_error(paths[1]) == 1
    assert forward_rp.pairs._dc == {}
    assert forward_rp.get_logical_error(paths[0] | paths[1]) == 2
    assert forward_rp.pairs._dc == {}


def test_get_logical_error_surface(sf_ol: Forward):

    d = sf_ol._CODE.D
    h = sf_ol._COMMIT_HEIGHT
    ts = [0, h-1]
    paths = [{((0, j, t), (0, j+1, t)) for j in range(-1, d-1)} for t in ts]
    assert sf_ol.get_logical_error(set()) == 0
    assert sf_ol.pairs._dc == {}
    assert sf_ol.get_logical_error(paths[0]) == 1
    assert sf_ol.pairs._dc == {}
    assert sf_ol.get_logical_error(paths[1]) == 1
    assert sf_ol.pairs._dc == {}
    assert sf_ol.get_logical_error(paths[0] | paths[1]) == 2
    assert sf_ol.pairs._dc == {}

    commit_leftover = {
        ((0, 0, h-1), (0, 0, h)),
        ((0, 0, h-1), (1, 0, h-1)),
        ((1, 0, h-1), (1, 0, h)),
    }
    assert sf_ol.get_logical_error(commit_leftover) == 0
    # test pair with zero j separation not ignored
    assert sf_ol.pairs._dc == {
        (0, 0, 0): (1, 0, 0),
        (1, 0, 0): (0, 0, 0),
    }


def test_get_syndrome(forward_rp: Forward):
    with mock.patch(
        "localuf._schemes.Forward._get_artificial_defects",
        return_value={(1, 0)},
    ) as mock_gad:
        next_syndrome = forward_rp._get_syndrome(set(forward_rp._CODE.EDGES), set())
        mock_gad.assert_called_once_with(set(forward_rp._CODE.EDGES))
        assert next_syndrome == {(1, 0)}


def test_get_artificial_defects(forward_rp: Forward):
    for j in range(forward_rp._CODE.D-1):
        commit_leftover = {((j, forward_rp._COMMIT_HEIGHT-1), (j, forward_rp._COMMIT_HEIGHT))}
        assert forward_rp._get_artificial_defects(commit_leftover) == {(j, 0)}
        commit_leftover = {((j, 0), (j, 1))}
        assert forward_rp._get_artificial_defects(commit_leftover) == set()


def test_get_artificial_defects_circuit_level(sfCL_OL_scheme: Forward):
    h = sfCL_OL_scheme._COMMIT_HEIGHT
    # test artificial defects from previous decoding cycle do not carry over
    commit_leftover: set[Edge] = {((0, -1, 0), (0, 0, 0))}  # this is a valid commit leftover!
    assert sfCL_OL_scheme._get_artificial_defects(commit_leftover) == set()
    # test SD edges considered (this could fail if we assumed v higher than u for each timelike uv)
    commit_leftover = {
        ((1, -1, h-1), (1, 0, h-1)),
        ((0, 0, h), (1, 0, h-1)),
    }
    assert sfCL_OL_scheme._get_artificial_defects(commit_leftover) == {(0, 0, 0)}
    # test artificial defect annihilation
    commit_leftover |= {
        ((0, -1, h-1), (0, 0, h-1)),
        ((0, 0, h-1), (0, 0, h)),
    }
    assert sfCL_OL_scheme._get_artificial_defects(commit_leftover) == set()


def test_make_error_in_buffer_region(forward_rp: Forward):
    noise_level = 0.5
    with mock.patch(
        "localuf.noise.Phenomenological.make_error",
        return_value=set(forward_rp._CODE.EDGES),
    ) as mock_me:
        error_in_buffer_region = forward_rp._make_error_in_buffer_region(noise_level)
        mock_me.assert_called_once_with(noise_level)
        assert error_in_buffer_region == set(forward_rp._BUFFER_EDGES)


class TestRun:

    @pytest.fixture(
            params=product(range(3, 7, 2), range(1, 4), range(1, 4)),
            ids=lambda x: f"d{x[0]} c{x[1]} b{x[2]}",
    )
    def forward(self, request):
        d, c, b = request.param
        return Repetition(
            d,
            'phenomenological',
            scheme='forward',
            commit_height=c,
            buffer_height=b,
        ).SCHEME
    
    def test_zero_cycles(self, forward: Forward):
        decoder = UF(forward._CODE)
        with pytest.raises(ValueError):
            forward.run(decoder, 1, 0)

    @pytest.mark.parametrize("n", range(1, 4), ids=lambda x: f"n{x}")
    def test_commit_equals_buffer(self, forward: Forward, n: int):
        uf = UF(forward._CODE)
        noise_level = 0.5
        cleanse_count = forward.WINDOW_HEIGHT // forward._COMMIT_HEIGHT
        commit_leftover = set()
        decoding_cycle_count = n+1 + cleanse_count
        with (
            mock.patch(
            "localuf._schemes.Forward._make_error_in_buffer_region",
            return_value=set()
        ) as mock_meibr,
            mock.patch("localuf._schemes.Forward._make_error") as mock_me,
            mock.patch("localuf._schemes.Forward._get_artificial_defects") as mock_gad,
            mock.patch("localuf.codes.Repetition.get_syndrome") as mock_gs,
            mock.patch("localuf.decoders.uf.UF.reset") as mock_reset,
            mock.patch("localuf.decoders.uf.UF.decode") as mock_decode,
            mock.patch(
            "localuf._schemes.Forward._get_leftover",
            return_value=(set(), set())
        ) as mock_gl,
            mock.patch(
            "localuf._schemes.Forward.get_logical_error",
            return_value=1
        ) as mock_gle,
        ):
            m, slenderness = forward.run(uf, noise_level, n)
            assert m == decoding_cycle_count
            layer_count = forward._BUFFER_HEIGHT + (n+1)*forward._COMMIT_HEIGHT
            assert slenderness == layer_count / forward._CODE.D
            mock_meibr.assert_called_once_with(noise_level)
            assert mock_me.call_args_list == n*[mock.call(set(), noise_level, exclude_future_boundary=False)] \
                + [mock.call(set(), noise_level, exclude_future_boundary=True)] \
                + cleanse_count * [mock.call(set(), 0, exclude_future_boundary=False)]
            assert mock_gad.call_args_list == [mock.call(commit_leftover)] * decoding_cycle_count
            assert mock_gs.call_args_list == [mock.call(mock_me.return_value)] * decoding_cycle_count
            syndrome = mock_gs.return_value ^ mock_gad.return_value
            assert mock_reset.call_args_list == [mock.call()] * decoding_cycle_count
            assert mock_decode.call_args_list == [mock.call(syndrome)] * decoding_cycle_count
            assert mock_gl.call_args_list == [mock.call(mock_me.return_value, uf.correction)] * decoding_cycle_count
            assert mock_gle.call_args_list == [mock.call(commit_leftover)] * decoding_cycle_count


def test_make_error(forward3F: Forward):
    # test commit made
    error = forward3F._make_error({
        ((0, 3), (0, 4)),
        ((0, 3), (1, 3)),
    }, 0)
    assert error == {
        ((0, 0), (0, 1)),
        ((0, 0), (1, 0)),
    }
    # test buffer made
    with mock.patch(
        "localuf.noise.Phenomenological.make_error",
        return_value=set(),
    ) as mock_me:
        error = forward3F._make_error(set(), 0.5)
        mock_me.assert_called_once_with(0.5)
    assert error == set()


import numpy as np

from localuf.noise import CircuitLevel

def test_make_multiplicities_both(
        assert_these_multiplicities_unchanged,
        diagonals,
):
    m = CircuitLevel._make_multiplicities(demolition=True, monolingual=True)
    assert (m['S'] == np.array((4, 2, 4, 0))).all()
    assert (m['E westmost'] == np.array((1, 0, 5, 0))).all()
    assert (m['E bulk'] == np.array((2, 0, 4, 0))).all()
    assert (m['E eastmost'] == np.array((1, 0, 4, 0))).all()
    assert (m['U 3'] == np.array((3, 0, 3, 2))).all()
    assert (m['U 4'] == np.array((4, 0, 2, 2))).all()
    assert_these_multiplicities_unchanged(m, diagonals)


from unittest import mock

import pytest

from localuf.decoders import Snowflake
from localuf.decoders.snowflake import _Node


@pytest.fixture
def sfn3a(snowflake3: Snowflake):
    """`_Node` instance with `NodeFriendship`."""
    return snowflake3.NODES[0, 1]


@pytest.mark.parametrize("prop", [
    "DROPEE",
])
def test_property_attributes(test_property, sfn3a: _Node, prop):
    test_property(sfn3a.FRIENDSHIP, prop)


def test_init(sfn3a: _Node):
    with mock.patch("localuf.decoders.snowflake.Friendship.__init__") as mock_init:
        sfn3a.FRIENDSHIP.__init__(sfn3a)
        mock_init.assert_called_once_with(sfn3a)


def test_drop(snowflake3: Snowflake):

    node_0 = snowflake3.NODES[0, 0]
    node_1 = snowflake3.NODES[0, 1]

    node_1.defect = True
    node_1.active = True

    with mock.patch("localuf.decoders.snowflake.Friendship.drop") as mock_drop:
        node_1.FRIENDSHIP.drop()
        mock_drop.assert_called_once_with()

    assert node_0.next_defect is True
    assert node_0.next_active is True


import pytest

from localuf.decoders.snowflake.main import _Repetition


def test_syndrome_vector_to_set():
    converter = _Repetition(3, 2)
    assert converter._syndrome_vector_to_set('00') == set()
    assert converter._syndrome_vector_to_set('10') == {(0, 1)}
    assert converter._syndrome_vector_to_set('01') == {(1, 1)}
    assert converter._syndrome_vector_to_set('11') == {(0, 1), (1, 1)}
    with pytest.raises(ValueError):
        converter._syndrome_vector_to_set('10010')


@pytest.mark.parametrize("window_height", [2, 3])
def test_set_to_syndrome_vector(window_height: int):
    converter = _Repetition(3, window_height)
    assert converter._set_to_syndrome_vector({(0, window_height-1), (1, window_height-1)}) == '11'
    assert converter._set_to_syndrome_vector(set()) == '00'


"""Unit test the `Controller` class.

These tests use a 2D distance-3 controller fixture.
Need not repeat on a 3D fixture.
"""
from unittest import mock

import pytest

from localuf import Surface
from localuf.decoders.luf import Controller, Macar
from localuf.decoders.luf.constants import Stage

def test_luf_attribute(c3: Controller):
    assert type(c3.LUF) is Macar

@pytest.mark.parametrize("prop", [
    "LUF",
])
def test_property_attributes(test_property, c3: Controller, prop):
    test_property(c3, prop)

def test_init(sf3T: Surface, c3: Controller):
    macar3T = Macar(sf3T)
    with mock.patch("localuf.decoders.luf.Controller.reset") as mock_reset:
        c3.__init__(macar3T)
        assert c3._LUF is macar3T
        mock_reset.assert_called_once()

def test_reset(c3: Controller):
    c3.stage = Stage.MERGING
    c3.reset()
    assert c3.stage is Stage.GROWING

def test_advance(c3: Controller):
    assert c3.advance()
    assert c3.stage is Stage.MERGING
    assert not c3.advance()
    assert c3.stage is Stage.PRESYNCING
    assert not c3.advance()
    assert c3.stage is Stage.SYNCING
    assert not c3.advance()
    assert c3.stage is Stage.BURNING
    assert c3.advance()
    assert c3.stage is Stage.PEELING
    assert c3.advance()
    assert c3.stage is Stage.DONE


from collections import Counter
from unittest import mock

import pytest

from localuf.decoders.luf import ActisNodes, ActisNode, NodeFriendship, Actis
from localuf.decoders.luf.constants import Stage

@pytest.fixture
def in_(actis_nodes: ActisNodes):
    v = (0, 0) if actis_nodes.LUF.CODE.DIMENSION==2 else (0, 0, 0)
    return ActisNode(actis_nodes, v)

@pytest.fixture(name="in3", params=[
    ("sf3F", "v00"),
    ("sf3T", "v000"),
])
def _in3(request):
    sf, v = (request.getfixturevalue(s) for s in request.param)
    actis = Actis(sf)
    ins = ActisNodes(actis)
    return ActisNode(ins, v)

@pytest.fixture
def in3F(sf3F, v00):
    actis = Actis(sf3F)
    ins = ActisNodes(actis)
    return ActisNode(ins, v00)

@pytest.fixture
def in3T(sf3T, v000):
    actis = Actis(sf3T)
    ins = ActisNodes(actis)
    return ActisNode(ins, v000)

@pytest.fixture
def in3_with_E(in3: ActisNode):
    """ActisNode (with correct access) and a pointer string."""
    pointer = 'E'
    e, index = in3.NEIGHBORS[pointer]
    in3.access = {pointer: in3.NODES.dc[e[index]]}
    return in3, pointer

def test_init(in_: ActisNode):

    d = in_.NODES.LUF.CODE.D
    em = in_.NODES.LUF.CODE.NOISE
    if str(em) == 'code capacity':
        assert in_.SPAN == 2 * (d-1)
    elif str(em) == 'phenomenological':
        assert in_.SPAN == 3 * (d-1)
    elif str(em) == 'circuit-level':
        assert in_.SPAN == d-1

    assert type(in_.FRIENDSHIP) is NodeFriendship

    with mock.patch("localuf.decoders.luf._Node.__init__") as mock_init:
        in_.__init__(in_.NODES, in_.INDEX)
        mock_init.assert_called_once_with(in_.NODES, in_.INDEX)


def test_countdown_start(actis_nodes: ActisNodes):
    d = actis_nodes.LUF.CODE.D
    counter = Counter(node.SPAN for node in actis_nodes.dc.values())
    if str(actis_nodes.LUF.CODE.NOISE) == 'code capacity':
        for cs in range(2*d):
            if cs < d:
                assert counter[cs] == cs + 1
            else:
                assert counter[cs] == 2*d - cs
    elif str(actis_nodes.LUF.CODE.NOISE) == 'phenomenological':
        # cs goes up somewhat as triangular numbers
        pass
    elif str(actis_nodes.LUF.CODE.NOISE) == 'circuit-level':
        for cs in range(1, d+1):
            small = d - cs
            assert counter[cs] == (small+1)**3 - small**3
        assert counter[0] == d**2


def test_advance(in_: ActisNode):
    # from https://stackoverflow.com/a/63690318/20887677
    with (
        mock.patch("localuf.decoders.luf.ActisNode.advance_definite") as definite_mock,
        mock.patch("localuf.decoders.luf.ActisNode.advance_indefinite") as indefinite_mock,
    ):
        in_.advance()
        definite_mock.assert_called_once_with('growing')
        in_.stage = Stage.MERGING
        in_.advance()
        indefinite_mock.assert_called_once_with('merging')
        in_.stage = Stage.PRESYNCING
        in_.advance()
        definite_mock.assert_called_with('presyncing')
        in_.stage = Stage.SYNCING
        in_.advance()
        indefinite_mock.assert_called_with('syncing')

def test_advance_definite(in_: ActisNode):
    with mock.patch("localuf.decoders.luf.ActisNode.growing") as growing_mock:
        in_.advance_definite('growing')

        growing_mock.assert_called_once_with()
    assert in_.next_stage is Stage.MERGING

    in_.countdown = 1
    in_.advance_definite('growing')

    assert in_.countdown == 0

def test_advance_indefinite(in_: ActisNode):
    in_.stage = Stage.MERGING
    with mock.patch(
        "localuf.decoders.luf.NodeFriendship.update_stage",
        return_value=True
    ) as us_mock:
        in_.advance_indefinite('merging')

        us_mock.assert_called_once_with()
        assert in_.countdown == in_.SPAN

    with (
        mock.patch(
            "localuf.decoders.luf.NodeFriendship.update_stage",
            return_value=False,
        ) as _,
        mock.patch("localuf.decoders.luf.ActisNode.merging") as merging_mock,
        mock.patch("localuf.decoders.luf.NodeFriendship.relay_signals") as rs_mock,
    ):
        in_.advance()

        merging_mock.assert_called_once_with()
        rs_mock.assert_called_once_with()


import pytest

from localuf.decoders import NodeBUF


@pytest.fixture
def dynamic_node_buf(sf7F):
    nbuf = NodeBUF(sf7F, dynamic=True)
    nbuf.history = []
    return nbuf


def test_validate_static(
        node_buf: NodeBUF,
        validated_static_erasure,
        syndrome7F,
):
    node_buf.validate(syndrome7F)
    assert len(node_buf.clusters) == 26
    assert all(bucket == set() for bucket in node_buf.buckets.values())
    assert node_buf.erasure == validated_static_erasure


def test_validate_dynamic(
        dynamic_node_buf: NodeBUF,
        validated_dynamic_erasure,
        syndrome7F,
):
    dynamic_node_buf.validate(syndrome7F)
    assert len(dynamic_node_buf.clusters) == 27
    assert all(bucket == set() for bucket in dynamic_node_buf.buckets.values())
    assert dynamic_node_buf.erasure == validated_dynamic_erasure


import pytest

from localuf.noise.main import _Uniform

class ConcreteUniform(_Uniform):
    """Instantiable version of `_Uniform`."""

    def __str__(self):
        return "concrete uniform"

@pytest.fixture
def uniform():
    return ConcreteUniform((
        ((0, 0), (0, 1)),
        ((0, 0), (1, 0)),
    ))


@pytest.mark.parametrize("prop", [
    "FRESH_EDGES",
])
def test_property_attributes(test_property, uniform: ConcreteUniform, prop):
    test_property(uniform, prop)


def test_make_error(uniform: ConcreteUniform):
    assert uniform.make_error(0) == set()
    assert uniform.make_error(1) == set(uniform.FRESH_EDGES)


def test_get_edge_weights(uniform: ConcreteUniform):
    flip_probability = 1e-1
    weight = uniform.log_odds_of_no_flip(flip_probability)
    assert uniform.get_edge_weights(flip_probability) == {e: (flip_probability, weight) for e in uniform.FRESH_EDGES}


def test_force_error(uniform: ConcreteUniform):
    assert uniform.force_error(0) == set()
    assert len(uniform.force_error(1)) == 1
    assert uniform.force_error(len(uniform.FRESH_EDGES)) == set(uniform.FRESH_EDGES)


from collections.abc import Callable
from itertools import product

import pytest

from localuf.sim._height_calculator import get_heights

d = 3
global_batch_slenderness = 100
products = tuple(product((None, lambda d: 2*(d//2), lambda d: 3*(d//2)), repeat=2))


@pytest.mark.parametrize('get_commit_height, get_buffer_height', products)
def test_batch(get_commit_height: None | Callable[[int], int], get_buffer_height: None | Callable[[int], int]):
    w, c, b = get_heights(d, global_batch_slenderness, 'batch', get_commit_height, get_buffer_height)
    assert w is None
    if get_commit_height is None:
        assert c is None
    else:
        assert c == get_commit_height(d)
    if get_buffer_height is None:
        assert b is None
    else:
        assert b == get_buffer_height(d)


@pytest.mark.parametrize('get_commit_height, get_buffer_height', products)
def test_global_batch(get_commit_height: None | Callable[[int], int], get_buffer_height: None | Callable[[int], int]):
    w, c, b = get_heights(d, global_batch_slenderness, 'global batch', get_commit_height, get_buffer_height)
    assert w == d*global_batch_slenderness
    if get_commit_height is None:
        assert c is None
    else:
        assert c == get_commit_height(d)
    if get_buffer_height is None:
        assert b is None
    else:
        assert b == get_buffer_height(d)


@pytest.mark.parametrize('get_commit_height, get_buffer_height', products)
def test_forward(get_commit_height: None | Callable[[int], int], get_buffer_height: None | Callable[[int], int]):
    w, c, b = get_heights(d, global_batch_slenderness, 'forward', get_commit_height, get_buffer_height)
    assert w is None
    if get_commit_height is None:
        assert c is d
    else:
        assert c == get_commit_height(d)
    if get_buffer_height is None:
        assert b is d
    else:
        assert b == get_buffer_height(d)


@pytest.mark.parametrize('get_commit_height, get_buffer_height', products)
def test_frugal(get_commit_height: None | Callable[[int], int], get_buffer_height: None | Callable[[int], int]):
    w, c, b = get_heights(d, global_batch_slenderness, 'frugal', get_commit_height, get_buffer_height)
    assert w is None
    if get_commit_height is None:
        assert c is 1
    else:
        assert c == get_commit_height(d)
    if get_buffer_height is None:
        assert b is 2*(d//2)
    else:
        assert b == get_buffer_height(d)


import pytest

from localuf.codes import Surface
from localuf.decoders.uf import UF, _Cluster


@pytest.fixture
def uf3T(sf3T: Surface):
    decoder = UF(sf3T)
    return decoder


@pytest.fixture
def two_clusters(uf3T: UF):
    east = (0, 2, 0)
    west = (0, -1, 0)
    larger = _Cluster(uf3T, east)
    smaller = _Cluster(uf3T, west)
    return larger, smaller


from unittest import mock

from localuf.decoders.luf import ActisNodes, UnoptimalWaiter
from localuf.decoders.luf.constants import Stage

def test_advance(actis_nodes: ActisNodes):
    actis_nodes._WAITER = UnoptimalWaiter(actis_nodes)
    iw = actis_nodes._WAITER
    with mock.patch("localuf.decoders.luf.ActisNodes.update_valid") as mock_update_valid:
        iw.advance()

        mock_update_valid.assert_not_called()
        assert actis_nodes.countdown == actis_nodes.SPAN + 1

        actis_nodes.countdown = 0
        actis_nodes.LUF.CONTROLLER.stage = Stage.MERGING
        iw.advance()

        mock_update_valid.assert_not_called()
        assert actis_nodes.countdown == actis_nodes.SPAN
        
        actis_nodes.countdown = 0
        actis_nodes.LUF.CONTROLLER.stage = Stage.PRESYNCING
        iw.advance()

        mock_update_valid.assert_not_called()
        assert actis_nodes.countdown == actis_nodes.SPAN

        actis_nodes.countdown = 0
        actis_nodes.LUF.CONTROLLER.stage = Stage.SYNCING
        iw.advance()

        mock_update_valid.assert_called_once_with()
        assert actis_nodes.countdown == actis_nodes.SPAN

    actis_nodes.busy_signal = True
    iw.advance()

    assert actis_nodes.countdown == actis_nodes.SPAN + 1

    actis_nodes.busy_signal = False
    iw.advance()

    assert actis_nodes.countdown == actis_nodes.SPAN


"""Classes for local Union--Find decoder."""

import abc
from functools import cached_property
from typing import Literal

import networkx as nx

from localuf.decoders.luf.constants import Stage, BUSY_SIGNAL_SYMBOLS, ACTIVE_SIGNAL_SYMBOLS
from localuf.type_aliases import Edge, Node
from localuf.codes import Surface
from localuf import noise
from localuf import constants
from localuf.constants import Growth
from localuf.decoders.policies import DigraphMaker, DecodeDrawer
from localuf.decoders._base_uf import direction, BaseUF


class LUF(BaseUF):
    """The graph used by local UF decoder.
    
    Extends ``_BaseUF``.
    
    Class attributes:
    
    * ``DEFAULT_X_OFFSET`` default ``x_offset``.
        Slightly larger than ``constants.DEFAULT_X_OFFSET``
        as the drawer for this class shows more information at each node.
    
    Instance attributes (1--5 constant):
    
    * ``CONTROLLER`` a global controller object.
    * ``NODES`` a Nodes object.
    * ``VISIBLE`` whether controller has direct connection w/ each node or only node w/ ID 0
        i.e. strictly local iff ``VISIBLE`` is ``False``.
        Only computed if ``syndrome`` not yet an attribute.
    * ``_DIGRAPH_MAKER`` maker of NetworkX digraph.
    * ``_DECODE_DRAWER`` provides ``draw_decode``.
    * ``_FIG_WIDTH`` figure width used by drawer.
    * ``correction`` only exists after calling ``decode()``.
    * ``_pointer_digraph`` a NetworkX digraph representing the fully grown edges used by pointers,
        the set of its edges as directed edges,
        the set of its edges as undirected edges.
    
    In drawings of this graph:
    
    * active nodes are square-shaped
    * inactive nodes are circular
    * CID is shown as a label
    * nodes with anyons are outlined in black
    * pointers are shown by arrows on edges
    * edges so far added to the correction are in red
    * the top-left node also shows the controller stage (PS stands for presyncing etc.)
    """

    DEFAULT_X_OFFSET = 0.3

    def __init__(self, code: Surface, visible=True, _optimal=True):
        """
        :param code: the code to be decoded.
        :param visible: whether controller has direct connection w/ each node or only node w/ ID 0
            i.e. strictly local iff ``visible`` is ``False``.
        :param _optimal: whether management of ``self.NODES.countdown`` is optimal.
            Relevant only when ``visible`` is ``False``.
        """
        self.correction = set()
        super().__init__(code)
        self._CONTROLLER = Controller(self)
        self._NODES = MacarNodes(self) if visible else ActisNodes(self, _optimal)
        self._VISIBLE = visible
        self._DIGRAPH_MAKER = DigraphMaker(self.NODES.dc, self.growth)
        self._DECODE_DRAWER = DecodeDrawer(self._FIG_WIDTH, fig_height=self._FIG_HEIGHT)

    @property
    def CONTROLLER(self): return self._CONTROLLER

    @property
    def NODES(self): return self._NODES

    @property
    def VISIBLE(self): return self._VISIBLE
    
    def decode(
            self,
            syndrome: set[Node],
            draw=False,
            log_history=False,
            **kwargs_for_draw_decode,
    ):
        """Additional inputs over those of ``Decoder.decode()``:
        
        * ``log_history`` whether to populate ``history`` attribute.
        * ``kwargs_for_draw_decode`` passed to ``self.draw_decode()``.
        
        
        :returns tSV: Number of timesteps to validate syndrome.
        :returns tBP: Number of timesteps to burn and peel.
        
                If ``log_history`` is ``True`` then ``tSV + tBP = len(self.history)``.
        """
        log_history |= draw
        tSV = self.validate(
            syndrome,
            draw=False,
            log_history=log_history,
        )
        tBP = self.peel(log_history)
        if draw:
            self.draw_decode(**kwargs_for_draw_decode)
        return tSV, tBP

    def validate(
            self,
            syndrome: set[Node],
            draw=False,
            log_history=False,
            **kwargs_for_draw_decode,
    ):
        """Validate syndrome.
        
        
        :param syndrome: the set of defects.
        :param draw: whether to draw.
        :param log_history: whether to populate ``history`` attribute.
        :param kwargs_for_draw_decode: passed to ``self.draw_decode()``.
        
        
        :returns: ``tSV`` # timesteps to validate syndrome. Equals ``len(self.history)`` if ``log_history`` is ``True``.
        """
        self.NODES.load(syndrome)
        tSV = 0
        log_history |= draw
        if log_history:
            self.init_history()
            while self.CONTROLLER.stage is not Stage.BURNING:
                self._advance()
                self.append_history()
                tSV += 1
        else:
            while self.CONTROLLER.stage is not Stage.BURNING:
                self._advance()
                tSV += 1
        if draw:
            self.draw_decode(**kwargs_for_draw_decode)
        return tSV

    def peel(self, log_history: bool):
        """Burn & peel after syndrome validation.
        
        
        ``log_history`` whether to populate ``history`` attribute.
        
        
        :returns: ``tBP`` # timesteps to burn and peel.
        """
        tBP = 0
        if log_history:
            while self.CONTROLLER.stage is not Stage.DONE:
                self._advance()
                self.append_history()
                tBP += 1
        else:
            while self.CONTROLLER.stage is not Stage.DONE:
                self._advance()
                tBP += 1
        return tBP

    @cached_property
    def syndrome(self):
        """Syndrome computed from ``self.NODES``."""
        return self.NODES.syndrome

    def reset(self):
        super().reset()
        self.CONTROLLER.reset()
        self.NODES.reset()
        try: del self.syndrome
        except AttributeError: pass
        try: del self.history
        except AttributeError: pass

    def _advance(self):
        """Advance 1 timestep."""
        self.NODES.advance()  # physical
        self.NODES.update_unphysicals()  # nonphysical
        growth_changed = self.CONTROLLER.advance()  # physical
        if growth_changed:  # nonphsyical
            self.NODES.update_access()

    # DRAWERS

    def draw_growth(
        self,
        highlighted_edges: set[Edge] | None = None,
        highlighted_edge_color='k',
        unhighlighted_edge_color=constants.DARK_GRAY,
        x_offset=DEFAULT_X_OFFSET,
        with_labels=True,
        labels: dict[Node, str] | None = None,
        show_global=True,
        node_size=constants.DEFAULT,
        linewidths=constants.WIDE_MEDIUM,
        anyon_color='k',
        active_shape='s',
        width=constants.WIDE_MEDIUM,
        arrows: bool | None = None,
        boundary_color=constants.BLUE,
        defect_color=constants.RED,
        nondefect_color=constants.GREEN,
        show_boundary_defects=True,
        defect_label_color='k',
        **kwargs_for_networkx_draw,
    ):
        g = self.CODE.GRAPH
        dig, dig_diedges, dig_edges = self._pointer_digraph
        pos = self.CODE.get_pos(x_offset)
        if highlighted_edges is None:
            highlighted_edges = self.correction
        if labels is None:
            labels = self.NODES.labels(show_global)
        nodes_w_anyons = {v for v, node in self.NODES.dc.items() if node.anyon}

        # DRAW INACTIVE NODES AND EDGES UNUSED BY POINTERS
        
        # node-related kwargs
        inactive_nodelist = [v for v, node in self.NODES.dc.items() if not node.active]
        inactive_node_color, inactive_edgecolors = \
            self._get_node_color_and_edgecolors(
                outlined_nodes=nodes_w_anyons,
                nodelist=inactive_nodelist,
                outline_color=anyon_color,
                boundary_color=boundary_color,
                defect_color=defect_color,
                nondefect_color=nondefect_color,
                show_boundary_defects=show_boundary_defects,
            )
        
        # edge-related kwargs
        edgelist = [
            e for e in self.CODE.EDGES
            if self.growth[e] in {Growth.HALF, Growth.FULL}
            and e not in dig_edges
        ] + list(highlighted_edges)
        edge_color, style = self._get_edge_color_and_style(
            edgelist,
            highlighted_edges,
            highlighted_edge_color,
            unhighlighted_edge_color,
        )
        nx.draw(
            g,
            pos=pos,
            nodelist=inactive_nodelist,
            node_size=node_size,
            node_color=inactive_node_color,
            linewidths=linewidths,
            edgecolors=inactive_edgecolors,
            edgelist=edgelist,
            width=width,
            edge_color=edge_color,
            style=style,
            **kwargs_for_networkx_draw,
        )

        # DRAW ACTIVE NODES
        active_nodelist = [v for v, node in self.NODES.dc.items() if node.active]
        active_node_color, active_edgecolors = \
            self._get_node_color_and_edgecolors(
                outlined_nodes=nodes_w_anyons,
                nodelist=active_nodelist,
                outline_color=anyon_color,
                boundary_color=boundary_color,
                defect_color=defect_color,
                nondefect_color=nondefect_color,
                show_boundary_defects=show_boundary_defects,
            )
            # need not actually pass `show_boundary_defects`
            # as boundary nodes are never active
            # but include in case need test correctness
        nx.draw_networkx_nodes(
            g,
            pos,
            nodelist=active_nodelist,
            node_size=node_size,
            node_color=active_node_color, # type: ignore
            node_shape=active_shape,
            linewidths=linewidths,
            edgecolors=active_edgecolors,
        )

        # DRAW EDGES USED BY POINTERS
        dig_edge_color = [
            highlighted_edge_color if e in highlighted_edges else
            unhighlighted_edge_color for e in dig_edges
        ]
        nx.draw_networkx_edges(
            dig,
            pos,
            node_size=node_size,
            edgelist=dig_diedges,
            width=width,
            edge_color=dig_edge_color, # type: ignore
            arrows=arrows,
        )

        if with_labels:
            # DRAW LABELS (defect labels `defect_label_color`; nondefect, black)
            defect_labels = {v: label for v, label in labels.items() if v in self.syndrome}
            nondefect_labels = {v: label for v, label in labels.items() if v not in self.syndrome}
            nx.draw_networkx_labels(g, pos, labels=defect_labels, font_color=defect_label_color)
            nx.draw_networkx_labels(g, pos, labels=nondefect_labels)


    @property
    def _pointer_digraph(self): return self._DIGRAPH_MAKER.pointer_digraph
    
    def draw_decode(self, **kwargs_for_networkx_draw):
        self._DECODE_DRAWER.draw(self.history, **kwargs_for_networkx_draw)

    @property
    def _FIG_WIDTH(self):
        return max(1, self.CODE.D * self._FIG_FACTOR)

    @property
    def _FIG_HEIGHT(self):
        return max(1, (self.CODE.D-1) * self._FIG_FACTOR)
    
    @property
    def _FIG_FACTOR(self):
        return 3*(self.CODE.DIMENSION-1) / (1+self.VISIBLE) / 2


class Controller:
    """Global controller for LUF.
    
    Instance attributes (1 constant):
    
    * ``LUF`` the LUF object which has this controller.
    * ``stage`` the global stage.
    """

    def __init__(self, luf: LUF) -> None:
        """Input: ``luf`` the LUF object which has this controller."""
        self._LUF = luf
        self.reset()

    @property
    def LUF(self): return self._LUF

    def reset(self):
        """Factory reset."""
        self.stage = Stage.GROWING

    def advance(self):
        """Advance 1 timestep.
        
        Output: ``growth_changed`` which is ``True``
        if any growths have changed
        (happens after growth and burning stages, and peeling steps);
        else, ``False``.
        
        Side effect: Update ``stage``.
        """
        growth_changed = self.stage is Stage.PEELING
        if not self.LUF.NODES.busy:  # stage complete
            if self.stage < Stage.SV_STAGE_COUNT:  # in syndrome validation
                if self.stage is Stage.SYNCING and self.LUF.NODES.valid:
                    # syndrome validation done
                    self.stage = Stage.BURNING
                else:
                    if self.stage is Stage.GROWING:
                        growth_changed = True
                    self.stage += Stage.INCREMENT
                    self.stage %= Stage.SV_STAGE_COUNT
            else:
                growth_changed = True
                self.stage += Stage.INCREMENT
        return growth_changed


class Nodes(abc.ABC):
    """Collection of all LUF nodes.
    
    Instance attributes (1--2 constant):
    
    * ``LUF`` the LUF object which has these nodes.
    * ``dc`` a dictionary where each
        key an index tuple;
        value, the node object at that index.
    * ``syndrome`` the set of defects.
    * ``busy`` is ``False`` iff we are sure all nodes are ready for next stage.
    * ``valid`` is ``True`` iff we are sure no node is active after presyncing stage
        i.e. ``syndrome`` has been validated.
    * ``_global_label`` the label of global information used by drawer.
    """
    
    def __init__(self, luf: LUF) -> None:
        """Input: ``luf`` the LUF object which has these nodes."""
        self._LUF = luf
        self.dc: dict[Node, _Node]

    @property
    def LUF(self): return self._LUF

    def load(self, syndrome: set[Node]):
        """Load syndrome onto nodes collection."""
        for v in syndrome:
            self.dc[v].make_defect()

    @property
    def syndrome(self):
        """Compute syndrome from nodes in ``dc``."""
        return {v for v, node in self.dc.items() if node.defect}

    def reset(self):
        """Factory reset."""
        for node in self.dc.values():
            node.reset()

    def advance(self):
        """Advance 1 timestep."""
        for node in self.dc.values():
            node.advance()

    def update_unphysicals(self):
        """After each merging or syncing step, set ``attribute`` := ``next_attribute`` for all nodes."""
        if self.LUF.CONTROLLER.stage is Stage.MERGING:
            for node in self.dc.values():
                node.update_after_merge_step()
        elif self.LUF.CONTROLLER.stage is Stage.SYNCING:
            for node in self.dc.values():
                node.update_after_sync_step()

    def update_access(self):
        """Update access for all nodes.
        
        Call after growth has changed.
        """
        for node in self.dc.values():
            node.update_access()
    
    def labels(self, show_global=True):
        """Return the labels dictionary for the drawer.
        
        
        ``show_global`` whether to prepend the global label to the top-left node label.
        
        
        :returns: ``result`` a dictionary where each key a node index as a tuple; value, the label for the node at that index.
        """
        result = {
            v: node.get_label()
            for v, node in self.dc.items()
        }
        if show_global:
            code = self.LUF.CODE
            top_left = (0, -1, code.D-1) if code.DIMENSION == 3 else (0, -1)
            result[top_left] = self._global_label + result[top_left]
        return result

    @property
    def _global_label(self):
        return str(self.LUF.CONTROLLER.stage)


class MacarNodes(Nodes):
    """Collection of all LUF nodes, where controller directly connects to each node.
    
    Extends ``Nodes``.
    """

    def __init__(self, luf: LUF) -> None:
        super().__init__(luf)
        self._dc = {v: MacarNode(nodes=self, index=v)
                    for v in luf.CODE.NODES}
        
    @property
    def dc(self): return self._dc

    @property
    def busy(self):
        return any(node.busy for node in self.dc.values())

    @property
    def valid(self):
        return not any(node.active for node in self.dc.values())


class ActisNodes(Nodes):
    """Collection of all LUF nodes, where controller only sees node whose ID is 0.
    
    Extends ``Nodes``.
    
    Additional instance attributes (1--2 constant):
    
    * ``SPAN`` distance from controller to furthest node (boundary node w/ highest ID).
    * ``WAITER`` object which decides how long controller must wait until
        it is sure no more info is being relayed (towards it).
    * ``countdown`` tracker used by ``WAITER``.
    * ``busy_signal`` whether collection receives a busy signal in current timestep.
    * ``active_signal`` whether collection has received an active signal during the current syncing stage.
    * ``next_{busy, active}_signal`` the provisional ``{busy, active}_signal`` for the next timestep.
    * ``{busy, active}_signal`` have string representations via ``[attribute]_signal_symbol`` property.
    
    Notes:
    This decoder starts in the GROWING stage.
    Alternatively, could start in the SYNCING stage and in ``load()``,
    call ``self.update_valid()`` instead of ``self.valid = False``.
    This would be beneficial if noise level low enough.
    """
    
    def __init__(self, luf: LUF, optimal=True):
        super().__init__(luf)
        self._dc = {v: ActisNode(nodes=self, index=v)
                    for v in luf.CODE.NODES}

        d = luf.CODE.D
        if isinstance(luf.CODE.NOISE, noise.CodeCapacity):
            self._SPAN = 2*d  # 1 + (d-1) + d
        elif isinstance(luf.CODE.NOISE, noise.Phenomenological):
            self._SPAN = 3*d - 1  # 1 + (d-1) + d + (d-1)
        else: # CircuitLevel
            self._SPAN = d + 1  # 1 + d
        self._WAITER = OptimalWaiter(self) if optimal else UnoptimalWaiter(self)

        self.reset()

    @property
    def dc(self): return self._dc

    @property
    def SPAN(self): return self._SPAN
    
    @property
    def WAITER(self): return self._WAITER
    
    @property
    def busy_signal_symbol(self):
        return BUSY_SIGNAL_SYMBOLS[self.busy_signal]

    @property
    def active_signal_symbol(self):
        return ACTIVE_SIGNAL_SYMBOLS[self.active_signal]

    def reset(self):
        self.busy = False
        self.valid = True
        self.countdown = 0
        self.busy_signal = False
        self.next_busy_signal = False
        self.active_signal = False
        self.next_active_signal = False
        super().reset()

    def load(self, syndrome):
        super().load(syndrome)
        self.valid = False

    def update_valid(self):
        """Update ``valid`` attribute. Call after finish SYNCING stage."""
        self.valid = not self.active_signal
        self.active_signal = False
        self.next_active_signal = False
    
    def advance(self):
        super().advance()
        self.WAITER.advance()

    def update_unphysicals(self):
        super().update_unphysicals()
        for node in self.dc.values():
            node.update_unphysicals_for_actis()
        self._update_unphysicals_for_actis()

    def _update_unphysicals_for_actis(self):
        """``next_{busy, active}_signal`` -> ``{busy, active}_signal``."""

        self.busy_signal = self.next_busy_signal
        self.next_busy_signal = False

        self.active_signal = self.next_active_signal

    @property
    def _global_label(self):
        label = super()._global_label
        label += f'{self.countdown}{self.busy_signal_symbol}{self.active_signal_symbol}\n'
        return label


class Waiter(abc.ABC):
    """Base class for ``WAITER`` attribute of ``ActisNodes`` instance.
    
    Instance attributes:
    
    * ``NODES`` the ``ActisNodes`` object the waiter belongs to.
    * ``RECEIVING_START`` the value the waiter sets countdown to
        if it receives a busy signal during the receiving window.
    """

    def __init__(self, nodes: ActisNodes) -> None:
        """Input: ``nodes`` the ``ActisNodes`` object the waiter belongs to."""
        self._NODES = nodes
        self.RECEIVING_START: int

    @property
    def NODES(self): return self._NODES

    @property
    @abc.abstractmethod
    def received_busy_signal(self) -> bool:
        """Whether to set ``countdown`` to ``RECEIVING_START``."""

    def advance(self):
        """Advance 1 timestep.
        
        Countdown start for controller stage:
        
        * GROWING or PRESYNCING is ``span`` as it takes ``span`` timesteps
            (``span-1`` .. -1 .. 0)
            for information to go from controller to furthest node.
        * MERGING is ``span+1`` as it takes ``span`` timesteps
            (``span`` .. -1 .. 1)
            for information to go from furthest node to controller
            [and furthest node (which is a boundary) can be busy].
        * SYNCING is ``span`` as it takes ``span-1`` timesteps
            (``span-1`` .. -1 .. 1)
            for information to go from furthest detector to controller
            (furthest boundary node never busy nor active).
        """
        self.NODES.busy = True
        if self.NODES.countdown == 0:  # stage complete
            self.NODES.busy = False
            self.NODES.countdown = self.NODES.SPAN
            match self.NODES.LUF.CONTROLLER.stage:
                case Stage.GROWING:
                    self.NODES.countdown += 1
                case Stage.SYNCING:
                    self.NODES.update_valid()
        # only ever happens when controller stage in {MERGING, SYNCING}
        elif self.received_busy_signal:
            self.NODES.countdown = self.RECEIVING_START
        else:
            self.NODES.countdown -= 1


class OptimalWaiter(Waiter):
    """Optimal waiter for ``ActisNodes``.
    
    Extends ``Waiter``.
    
    Additional class constants:
    
    * ``RECEIVING_WINDOW`` the countdown values
        during which the waiter considers busy signals.
    """

    RECEIVING_WINDOW = {1, 2}
    RECEIVING_START = 2

    @property
    def received_busy_signal(self):
        return self.NODES.busy_signal and self.NODES.countdown in self.RECEIVING_WINDOW


class UnoptimalWaiter(Waiter):
    """Unoptimal waiter for ``ActisNodes``.
    
    Extends ``Waiter``.
    
    Receiving window is the entire countdown i.e. always.
    """
    
    def __init__(self, nodes: ActisNodes):
        super().__init__(nodes)
        self.RECEIVING_START = nodes.SPAN + 1

    @property
    def received_busy_signal(self):
        return self.NODES.busy_signal


class _Node(abc.ABC):
    """Node for LUF.
    
    Class attributes:
    
    * ``OPPOSITE`` a dictionary where each
        key a possible ``pointer`` value;
        value, the value in the opposite direction.
    
    Instance attributes (1--4 constant,
    5 constant within a decoding cycle,
    6--14 variable within a decoding cycle):
    
    * ``NODES`` the Nodes object the node belongs to.
    * ``INDEX`` the index of the node.
    * ``ID`` the unique ID of the node.
        ``ID`` bijects ``INDEX``.
    * ``NEIGHBORS`` a dictionary where each
        key a pointer string;
        value, a tuple of the form (edge, index of edge which is neighbor).
    * ``_IS_BOUNDARY`` whether node is a boundary node.
    * ``defect`` whether the node is a defect.
    * ``active`` whether the cluster the node is in is active.
    * ``cid`` the ID of the cluster the node belongs to.
    * ``next_cid`` the provisional ``cid`` for the next timestep.
        It depends on the ``cid`` of the nodes around it.
        At the end of the timestep, each node sets ``cid = next_cid``.
        This ensures causality i.e. info travels at 1 edge per timestep.
    * ``anyon`` whether the node has an anyon.
    * ``next_anyon`` the provisional ``anyon`` for the next timestep.
    * ``pointer`` the direction the node sends anyons along.
    * ``busy`` whether the node changes any of its attributes in the current timestep.
    * ``access`` refers to neighbors along fully grown edges.
        In the form of a dictionary of where each
        key a pointer string;
        value, the ``_Node`` object in the direction of that pointer.
    """

    OPPOSITE: dict[direction, direction] = {
        'N': 'S',
        'W': 'E',
        'E': 'W',
        'S': 'N',
        'D': 'U',
        'U': 'D',
        'SD': 'NU',
        'NU': 'SD',
        'EU': 'WD',
        'WD': 'EU',
        'SEU': 'NWD',
        'NWD': 'SEU',
    }

    def __init__(self, nodes: Nodes, index: Node) -> None:
        """
        :param nodes: the Nodes object the node belongs to.
        :param index: the index of the node.
        """
        self._NODES = nodes
        self._INDEX = index
        self._ID = nodes.LUF.CODE.index_to_id(index)
        provisional_neighbors: tuple[tuple[direction, Edge, Literal[0, 1]], ...]
        """During merging,
        node will look at neighbors in order of this list,
        so order matters!
        """
        if isinstance(nodes.LUF.CODE.NOISE, noise.CodeCapacity):
            i, j = index
            provisional_neighbors = (
                ('N', ((i-1, j), index), 0),
                ('W', ((i, j-1), index), 0),
                ('E', (index, (i, j+1)), 1),
                ('S', (index, (i+1, j)), 1),
            )
        else:
            i, j, t = index
            n = ('N', ((i-1, j, t), index), 0)
            w = ('W', ((i, j-1, t), index), 0)
            d = ('D', ((i, j, t-1), index), 0)
            u = ('U', (index, (i, j, t+1)), 1)
            e = ('E', (index, (i, j+1, t)), 1)
            s = ('S', (index, (i+1, j, t)), 1)
            if isinstance(nodes.LUF.CODE.NOISE, noise.Phenomenological):
                provisional_neighbors = (n, w, d, u, e, s)
            else: # 'circuit-level'
                provisional_neighbors = (
                    ('NWD', ((i-1, j-1, t-1), index), 0),
                    n,
                    ('NU', ((i-1, j, t+1), index), 0),
                    ('WD', ((i, j-1, t-1), index), 0),
                    w, d, u, e,
                    ('EU', (index, (i, j+1, t+1)), 1),
                    ('SD', (index, (i+1, j, t-1)), 1),
                    s,
                    ('SEU', (index, (i+1, j+1, t+1)), 1),
                )
        # match case using CODE.INCIDENT_EDGES?
        self._NEIGHBORS: dict[direction, tuple[Edge, Literal[0, 1]]] = {
            pointer: (e, e_index)
            for pointer, e, e_index in provisional_neighbors
            if e in nodes.LUF.CODE.INCIDENT_EDGES[index]
        }
        self._IS_BOUNDARY = nodes.LUF.CODE.is_boundary(index)
        self.reset()

    @property
    def NODES(self): return self._NODES

    @property
    def INDEX(self): return self._INDEX

    @property
    def ID(self): return self._ID

    @property
    def NEIGHBORS(self): return self._NEIGHBORS

    def reset(self):
        """Factory reset."""
        self.defect = False
        self.active = False
        self.cid = self.ID
        self.next_cid = self.ID
        self.anyon = False
        self.next_anyon = False
        self.pointer: direction = 'C'
        self.busy = False
        self.access: dict[direction, _Node] = {}

    def make_defect(self):
        """Make the node a defect."""
        self.defect = True
        self.active = True
        self.anyon = True

    def growing(self):
        """Growth-increment all active incident edges."""
        if self.active:
            g = self.NODES.LUF
            edges_to_grow = {
                e for e in g.CODE.INCIDENT_EDGES[self.INDEX]
                if g.growth[e] in g._ACTIVE_GROWTH_VALUES
            }
            for e in edges_to_grow:
                g.growth[e] += Growth.INCREMENT

    def update_access(self):
        """Update access.
        
        Call after growth has changed.
        """
        self.access = {
            pointer: self.NODES.dc[e[index]]
            for pointer, (e, index) in self.NEIGHBORS.items()
            if self.NODES.LUF.growth[e] is Growth.FULL
        }

    def merging(self):
        """Relay anyon, and update CID and pointer depending on access."""
        self.busy = False
        if all((
            not self._IS_BOUNDARY,
            self.pointer != 'C',
            self.anyon,
        )):  # relay along pointer
            self.busy = True
            self.access[self.pointer].next_anyon ^= True
            self.anyon = False
        for pointer, neighbor in self.access.items():
            if neighbor.cid < self.next_cid:
                self.busy = True
                self.pointer = pointer
                self.next_cid = neighbor.cid

    def presyncing(self):
        """Initialise for syncing stage i.e. update after merge stage.
        
        If node the root of an active cluster, set ``active = True``;
        else, set ``active = False``.
        """
        self.active = all((
            not self._IS_BOUNDARY,
            self.pointer == 'C',
            self.anyon,
        ))

    @abc.abstractmethod
    def advance(self):
        """Advance 1 timestep."""

    def update_after_merge_step(self):
        """``next_{cid, anyon}`` -> ``{cid, anyon}``."""
        self.cid = self.next_cid
        # if replace ^= by = then a nonroot boundary node w/ anyon
        # will incorrectly lose its anyon in next timestep
        self.anyon ^= self.next_anyon
        self.next_anyon = False
    
    def syncing(self):
        """Update ``active`` depending on access."""
        self.busy = not self.active and any(
            neighbor.active for neighbor in self.access.values()
        )

    def update_after_sync_step(self):
        """``active`` := ``active`` OR ``busy``."""
        self.active |= self.busy

    def burning(self):
        """Burn edges whose endpoints point not along edge."""
        owned_edges: dict[direction, tuple[Edge, Literal[1]]] = {pointer: (e, index)
                       for pointer, (e, index) in self.NEIGHBORS.items()
                       if index == 1}
        for pointer, (e, index) in owned_edges.items():
            if self.NODES.LUF.growth[e] is Growth.FULL:
                if self.pointer != pointer:
                    if self.NODES.dc[e[index]].pointer != self.OPPOSITE[pointer]:
                        self.NODES.LUF.growth[e] = Growth.UNGROWN

    def peeling(self):
        """Peel self if leaf.
        
        Note the root must not be peeled
        even if it is incident to exactly one fully grown edge.
        """
        self.busy = False
        if self.pointer != 'C' and len(self.access) == 1:
            self.busy = True
            pointer, neighbor = next(iter(self.access.items()))
            e, _ = self.NEIGHBORS[pointer]
            self.NODES.LUF.growth[e] = Growth.UNGROWN
            if self.defect:
                self.defect = False
                self.NODES.LUF.correction.add(e)
                neighbor.defect ^= True
    
    def get_label(self):
        """Return node information for drawer."""
        return str(self.cid)


class MacarNode(_Node):
    """Node for LUF which directly accesses global controller.
    
    Extends ``_Node``.
    """

    def advance(self):
        match self.NODES.LUF.CONTROLLER.stage:
            case Stage.GROWING:
                self.growing()
            case Stage.MERGING:
                self.merging()
            case Stage.PRESYNCING:
                self.presyncing()
            case Stage.SYNCING:
                self.syncing()
            case Stage.BURNING:
                self.burning()
            case Stage.PEELING:
                self.peeling()


class ActisNode(_Node):
    """Node for LUF w/o direct access to blind global controller.
    
    Extends ``_Node``.
    
    Additional instance attributes (1--2 constant):
    
    * ``SPAN`` the countdown start of the node.
    * ``FRIENDSHIP`` the type of connection the node has for communicating
        ``stage`` and signal information.
    * ``countdown`` tracks how long node must wait until staging is done
        i.e. when all nodes have same stage, when ``stage`` in {GROWING, PRESYNCING}.
    * ``stage`` node stage.
    * ``busy_signal`` if in the current timestep the node either
        is busy
        or receives a busy signal from another node.
    * ``active_signal`` if in the current timestep the node either
        has ``stage`` as ``presyncing`` and becomes active,
        or receives an active signal from another node.
    * ``next_{stage, {busy, active}_signal}`` the provisional ``{stage, {busy, active}_signal}`` for the next timestep.
    * ``stage, {busy, active}_signal`` have string representations via ``[attribute]_symbol`` property.
    """

    def __init__(self, nodes: ActisNodes, index: Node) -> None:

        d = nodes.LUF.CODE.D
        if isinstance(nodes.LUF.CODE.NOISE, noise.CircuitLevel):
            i, j, t = index
            depth = max(i, j+1, t)  # distance to node 0
            self._SPAN = d - depth  # `d` the staging & signalling tree height
        else:  # CodeCapacity or Phenomenological
            self._SPAN = nodes.LUF.CODE.DIMENSION * (d-1) - sum(index)

        super().__init__(nodes, index)
        self._FRIENDSHIP = ControllerFriendship(self) if self.ID == 0 else NodeFriendship(self)

    @property
    def NODES(self):
        self._NODES: ActisNodes
        return self._NODES

    @property
    def SPAN(self): return self._SPAN

    @property
    def FRIENDSHIP(self): return self._FRIENDSHIP

    @property
    def stage_symbol(self):
        return str(self.stage) if self.stage is not self.NODES.LUF.CONTROLLER.stage else ''

    @property
    def busy_signal_symbol(self):
        return BUSY_SIGNAL_SYMBOLS[self.busy_signal]

    @property
    def active_signal_symbol(self):
        return ACTIVE_SIGNAL_SYMBOLS[self.active_signal]

    def reset(self):
        super().reset()
        self.countdown = 0
        self.stage = Stage.GROWING
        self.next_stage = self.stage
        self.busy_signal = False
        self.next_busy_signal = False
        self.active_signal = False
        self.next_active_signal = False

    def advance(self):
        match self.stage:
            case Stage.GROWING:
                self.advance_definite('growing')
            case Stage.MERGING:
                self.advance_indefinite('merging')
            case Stage.PRESYNCING:
                self.advance_definite('presyncing')
            case Stage.SYNCING:
                self.advance_indefinite('syncing')

    def advance_definite(self, stage: str):
        """Advance when ``stage`` in ``{'growing', 'presyncing'}``."""
        if self.countdown == 0:
            getattr(self, stage)()
            self.next_stage += Stage.INCREMENT
        else:
            self.countdown -= 1

    def advance_indefinite(self, stage: str):
        """Advance when ``stage`` in ``{'merging', 'syncing'}``."""
        changed = self.FRIENDSHIP.update_stage()
        if changed:
            self.countdown = self.SPAN
        else:
            getattr(self, stage)()
            self.FRIENDSHIP.relay_signals()

    def presyncing(self):
        """Initialise for syncing stage i.e. update after merge stage.
        
        If node the root of an active cluster, set ``active``, ``active_signal`` to ``True``;
        else, set both to ``False``.
        """
        super().presyncing()
        self.next_active_signal = self.active

    def update_unphysicals_for_actis(self):
        """``next_{stage, {busy, active}_signal}`` -> ``stage, {busy, active}_signal``."""
        self.stage = self.next_stage

        self.busy_signal = self.next_busy_signal
        self.next_busy_signal = False

        self.active_signal = self.next_active_signal
        self.next_active_signal = False

    def get_label(self):
        north = super().get_label()
        sw = self.countdown if self.stage in {
            Stage.GROWING,
            Stage.PRESYNCING,
        } else self.busy_signal_symbol
        label = f'{north}\n{self.stage_symbol}{sw} {self.active_signal_symbol}'
        return label


class Friendship(abc.ABC):
    """The type of connection for communicating ``stage, {busy, active}_signal`` information.
    
    Instance attributes (1 constant):
    
    * ``NODE`` the node which has this friendship.
    """

    def __init__(self, node: ActisNode) -> None:
        """Input: ``node`` node which has this friendship."""
        self._NODE = node

    @property
    def NODE(self): return self._NODE

    @abc.abstractmethod
    def update_stage(self) -> bool:
        """Update stage depending on neighbors and return whether stage changed."""

    def update_stage_helper(self, relayee: Controller | ActisNode):
        """Update stage based on stage of component ``relayee``."""
        changed = False
        # could just check if self.NODE.next_stage == relayee.stage?
        if (
            self.NODE.next_stage is Stage.MERGING and
            (rs:=relayee.stage) is Stage.PRESYNCING
        ) or (
            self.NODE.next_stage is Stage.SYNCING and
            (rs:=relayee.stage) is Stage.GROWING
        ):
            changed = True
            self.NODE.next_stage = rs
        return changed

    @abc.abstractmethod
    def relay_signals(self) -> None:
        """Relay ``{busy, active}_signal`` toward controller."""


class ControllerFriendship(Friendship):
    """Friendship w/ controller.
    
    Extends ``Friendship``.
    Only node whose ID is 0 has this friendship.
    Note this node never busy.
    """

    def update_stage(self):
        changed = self.update_stage_helper(self.NODE.NODES.LUF.CONTROLLER)
        return changed

    def relay_signals(self):
        self.NODE.NODES.next_busy_signal = self.NODE.busy_signal
        self.NODE.NODES.next_active_signal |= self.NODE.active_signal


class NodeFriendship(Friendship):
    """Friendship w/ neighboring nodes.
    
    Extends ``Friendship``.
    
    Additional instance attributes (1 constant):
    
    * ``RELAYEE`` the neighboring node that ``busy_signal`` is sent to.
        Equals the neighbor in the direction toward node of ID 0.
    """

    def __init__(self, node: ActisNode) -> None:
        if isinstance(node.NODES.LUF.CODE.NOISE, noise.CircuitLevel):
            i, j, t = node.INDEX
            self._RELAYEE = (
                i-1 if i > 0 else i,
                j-1 if j >-1 else j,
                t-1 if t > 0 else t,
            )
        else:  # CodeCapacity or Phenomenological
            match node.INDEX:
                case i, j:
                    if j > -1:
                        self._RELAYEE = i, j-1
                    elif i > 0:
                        self._RELAYEE = i-1, j
                    else:
                        self._RELAYEE = 0, -2
                case i, j, t:
                    if t > 0:
                        self._RELAYEE = i, j, t-1
                    elif j > -1:
                        self._RELAYEE = i, j-1, t
                    elif i > 0:
                        self._RELAYEE = i-1, j, t
                    else:
                        self._RELAYEE = 0, -2, 0
        super().__init__(node)

    @property
    def RELAYEE(self) -> Node: return self._RELAYEE

    def update_stage(self):
        changed = self.update_stage_helper(self.NODE.NODES.dc[self.RELAYEE])
        return changed

    def relay_signals(self):
        self.NODE.busy_signal |= self.NODE.busy
        self.NODE.NODES.dc[self.RELAYEE].next_busy_signal |= self.NODE.busy_signal
        self.NODE.NODES.dc[self.RELAYEE].next_active_signal |= self.NODE.active_signal




from __future__ import annotations

import os
import sys
from datetime import date
from pathlib import Path


# Ensure the package is importable when building docs.
REPO_ROOT = Path(__file__).resolve().parents[1]
sys.path.insert(0, str(REPO_ROOT))


project = "localuf"
copyright = f"{date.today().year}"  # noqa: A001
author = "Tim Chan"


extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
]


autosummary_generate = True

autodoc_default_options = {
    "members": True,
    "undoc-members": False,
    "show-inheritance": True,
}
autodoc_member_order = "bysource"
autodoc_typehints = "description"


templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


html_theme = "alabaster"
html_static_path = ["_static"]


# If your docs environment does not have all runtime deps installed,
# you can add module names here to avoid import failures during autodoc.
autodoc_mock_imports: list[str] = []


# Be forgiving about missing references by default.
nitpicky = False


# Make sure Sphinx doesn't treat warnings as errors unless user opts in.
os.environ.setdefault("SPHINX_DISABLE_PROMPT", "1")



from unittest import mock

import pytest

from localuf.decoders import Snowflake
from localuf.decoders.snowflake import _Node


@pytest.fixture
def sfn3b(snowflake3: Snowflake):
    """`_Node` instance with `TopSheetFriendship`."""
    h = snowflake3.CODE.SCHEME.WINDOW_HEIGHT
    return snowflake3.NODES[0, h-1]


def test_drop(sfn3b: _Node):
    sfn3b.next_defect = True
    sfn3b.next_active = True
    with mock.patch("localuf.decoders.snowflake.NodeFriendship.drop") as mock_drop:
        sfn3b.FRIENDSHIP.drop()
        mock_drop.assert_called_once_with()
    assert sfn3b.next_defect is False
    assert sfn3b.next_active is False


"""Module for constants."""

from enum import IntEnum

from scipy.stats import norm

RED, GREEN, BLUE = '#Ff6666', '#Bef4be', '#9cdaed'
DARK_GRAY, GRAY, LIGHT_GRAY = '#b6b6b6', '#c6cbd3', '#ecedf0'
V_WIDE, WIDE, WIDE_MEDIUM, MEDIUM, MEDIUM_THIN, THIN = 5.0, 3.0, 2.0, 1.5, 1.0, 0.5
DEFAULT, SMALL = 300, 50
DEFAULT_X_OFFSET, STREAM_X_OFFSET = 0.2, 0.25

STANDARD_ERROR_ALPHA: float = 2 * (1-norm.cdf(1)) # type: ignore
"""Significance level corresponding to 1 SE. Roughly 32%."""


class Growth(IntEnum):
    """Constants for growth values."""

    BURNT, UNGROWN, HALF, FULL = range(-1, 3)
    INCREMENT = 1

    def __iadd__(self, other: 'Growth'):
        return Growth(self.value + other.value)
    
    @property
    def as_float(self) -> float:
        """Convert growth value to float. Burnt takes the value of 1."""
        if self is Growth.BURNT:
            return 1
        return self.value/2


import itertools

import pytest

import localuf
from localuf import Repetition, Surface

@pytest.fixture(name="sf", params=itertools.product(
    range(3, 11, 2),
    (
        'code capacity',
        'phenomenological',
        'circuit-level',
    ),
), ids=lambda x: f"d{x[0]} {x[1]}")
def _sf(request):
    return localuf.Surface(*request.param)

@pytest.fixture
def sf3F():
    return localuf.Surface(3, 'code capacity')

@pytest.fixture
def sf3T():
    return localuf.Surface(3, 'phenomenological')

@pytest.fixture  # unused
def sf3O():
    return localuf.Surface(
        d=3,
        noise='phenomenological',
        scheme='forward',
    )

@pytest.fixture
def sf5F():
    return localuf.Surface(5, 'code capacity')

@pytest.fixture
def sf5T():
    return localuf.Surface(5, 'phenomenological')

@pytest.fixture
def sf7F():
    return localuf.Surface(7, 'code capacity')

@pytest.fixture
def sf9F():
    return localuf.Surface(9, 'code capacity')

@pytest.fixture
def test_property():
    def f(instance: object, prop: str):
        with pytest.raises(AttributeError):
            delattr(instance, prop)
        with pytest.raises(AttributeError):
            setattr(instance, prop, None)
    return f


@pytest.fixture
def rp3_forward():
    return Repetition(
        3,
        noise='phenomenological',
        scheme='forward',
    )


@pytest.fixture
def rp3_frugal():
    return Repetition(
        3,
        noise='phenomenological',
        scheme='frugal',
    )


@pytest.fixture
def batch3F(sf3F: Surface):
    return sf3F.SCHEME


@pytest.fixture
def forward_rp(rp_forward: Repetition):
    return rp_forward.SCHEME


@pytest.fixture(params=range(3, 11, 2), ids=lambda x: f"d{x}")
def rp_forward(request):
    return Repetition(
        request.param,
        noise='phenomenological',
        scheme='forward',
    )


@pytest.fixture(
        name="sfCL",
        params=itertools.product(range(3, 9, 2), range(2, 5)),
        ids=lambda x: f"d{x[0]} h{x[1]}",
)
def _sfCL(request):
    d, h = request.param
    return Surface(d=d, noise='circuit-level', window_height=h)


@pytest.fixture(
        name="sfCL_OL",
        params=range(3, 9, 2),
        ids=lambda x: f"d{x}",
)
def _sfCL_OL(request):
    return Surface(
        d=request.param,
        noise='circuit-level',
        scheme='forward',
    )

@pytest.fixture
def sfCL_OL_scheme(sfCL_OL: Surface):
    return sfCL_OL.SCHEME


"""Emulate logical success/failure to deduce accuracy thresholds.

Available functions:

* monte_carlo
* monte_carlo_pymatching
* monte_carlo_special
* subset_sample
"""

from collections.abc import Callable
import itertools
from typing import Sequence, Type

import numpy as np
import pandas as pd
from pandas import DataFrame, MultiIndex

from localuf.sim._height_calculator import get_heights
from localuf.type_aliases import NoiseModel, DecodingScheme, Parametrization, FloatSequence, IntSequence
from localuf._base_classes import Code, Decoder
from localuf.decoders import MWPM


def monte_carlo(
        sample_counts: dict[int, list[tuple[float, int]]],
        code_class: Type[Code],
        decoder_class: Type[Decoder],
        noise: NoiseModel,
        scheme: DecodingScheme = 'batch',
        get_commit_height: Callable[[int], int] | None = None,
        get_buffer_height: Callable[[int], int] | None = None,
        parametrization: Parametrization = 'balanced',
        demolition: bool = False,
        monolingual: bool = False,
        _merge_redundant_edges: bool = True,
        **kwargs_for_decoder_class,
):
    """Make threshold data for any decoder.
    
    :param sample_counts: A dictionary mapping code distances to lists of (noise level, sample count) pairs.
        If ``scheme`` is 'global batch',
        each sample count must be the same for all noise levels of a given distance.
        For a more detailed definition of 'sample', see ``_base_classes.Scheme.run``.
    :param code_class: The class of the code; either ``localuf.Repetition`` or ``localuf.Surface``.
    :param decoder_class: The class of the decoder;
        available decoders are listed in ``localuf.decoders``.
    :param noise: The noise model.
    :param scheme: The decoding scheme; either 'batch', 'global batch', 'forward' or 'frugal'.
        In global batch scheme, the decoding graph is ``d*n`` layers tall i.e.
        as tall as the (entire) decoding graph would be in forward scheme with commit height ``d``.
    :param get_commit_height: A function with input ``d`` that outputs commit height
        e.g. ``lambda d: 2*(d//2)``.
        If ``None``, commit height is ``d`` for forward scheme and ``1`` for frugal scheme.
        Affects only forward and frugal decoding schemes.
    :param get_buffer_height: A function with input ``d`` that outputs buffer height.
        If ``None``, buffer height is ``d`` for forward scheme and ``2*(d//2)`` for frugal scheme.
        Affects only forward and frugal decoding schemes.
    :param parametrization: defines relative fault probabilities of
        1- and 2-qubit gates, and prep/measurement.
        Affects only circuit-level noise.
    :param demolition: whether measurement destroys the ancilla qubit state
        which hence needs to be initialized for next measurement cycle.
        Affects only circuit-level noise.
    :param monolingual: whether can prep/measure in only Z basis
        hence X-basis prep/measurement needs Hadamard gates.
        Affects only circuit-level noise.
    :param _merge_redundant_edges: whether to merge redundant boundary edges.
        Affects only circuit-level noise.
    :param kwargs_for_decoder_class: are for ``decoder_class``.
    
    :return df: A DataFrame where each column is a (distance, probability);
        rows 'm', 'n' indicate number of logical errors and samples, respectively.
    """
    dc: dict[tuple[int, float], tuple[int, int | float]] = {}
    for d, list_ in sample_counts.items():
        window_height, commit_height, buffer_height = get_heights(
            d,
            list_[0][1],
            scheme,
            get_commit_height,
            get_buffer_height,
        )
        code = code_class(
            d,
            noise,
            scheme=scheme,
            window_height=window_height,
            commit_height=commit_height,
            buffer_height=buffer_height,
            parametrization=parametrization,
            demolition=demolition,
            monolingual=monolingual,
            _merge_redundant_edges=_merge_redundant_edges,
        )
        decoder = decoder_class(code, **kwargs_for_decoder_class)
        for noise_level, n in list_:
            dc[d, noise_level] = code.SCHEME.run(decoder, noise_level, n)
    df = DataFrame(dc, index=('m', 'n'))
    df.columns.set_names(['d', 'p'], inplace=True)
    return df


def monte_carlo_results_to_sample_counts(results: DataFrame):
    """Convert ``monte_carlo`` output to its first input.
    
    Assumes row 'n' of ``results`` is the input of ``_base_classes.Scheme.run``.
    """
    sample_counts: dict[int, list[tuple[float, int]]] = {}
    for (d, noise_level), (_, n) in results.items(): # type: ignore
        if d not in sample_counts:
            sample_counts[d] = []
        sample_counts[d].append((noise_level, int(n)))
    return sample_counts


def monte_carlo_pymatching(
        ds: Sequence[int],
        noise_levels: FloatSequence,
        ns: int | IntSequence,
        code_class: Type[Code],
        noise: NoiseModel,
        **kwargs_for_code_class,
):
    """Make threshold data for PyMatching decoder.
    
    
    :param ds: same as for ``monte_carlo``.
    :param noise_levels: same as for ``monte_carlo``.
    :param ns: same as for ``monte_carlo``.
    :param code_class: same as for ``monte_carlo``.
    :param noise: same as for ``monte_carlo``.
    :param kwargs_for_code_class: for ``code_class``.
    
    :return df: Output same as for ``monte_carlo``.
    """
    if isinstance(ns, int):
        ns = [ns] * len(noise_levels)
    elif len(ns) != len(noise_levels):
        raise ValueError('Sequence `ns` must have same length as `ps`')
    mi = MultiIndex.from_product([list(ds), list(noise_levels)], names=['d', 'p'])
    df = DataFrame(columns=mi)
    for d in ds:
        code = code_class(
            d=d,
            noise=noise,
            **kwargs_for_code_class,
        )
        decoder = MWPM(code)
        dc: dict[float, tuple[int, int]] = {}
        for noise_level, n in zip(noise_levels, ns):
            matching = decoder.get_matching(noise_level)
            m = 0
            for _ in itertools.repeat(None, n):  # TODO: use `decode_batch`
                error, syndrome = matching.add_noise() # type: ignore
                correction = matching.decode(syndrome)
                m += not np.array_equal(error, correction) # type: ignore
            dc[noise_level] = (m, n)
        df[d] = DataFrame(dc, index=('m', 'n'))
    return df

def monte_carlo_special(
        ds: Sequence[int],
        noise_levels: FloatSequence,
        ns: int | IntSequence,
        code_class: Type[Code],
        decoder_class: Type[Decoder],
        parametrization: Parametrization = 'balanced',
        demolition: bool = False,
        monolingual: bool = False,
        _merge_redundant_edges: bool = True,
        **kwargs_for_decoder_class,
):
    """Make threshold data where
    code noise model is circuit-level;
    decoder noise model, phenomenological.
    
    Input & output same as for ``monte_carlo``.
    """
    if isinstance(ns, int):
        ns = [ns] * len(noise_levels)
    elif len(ns) != len(noise_levels):
        raise ValueError('Sequence ns must have same length as noise_levels')
    mi = MultiIndex.from_product([list(ds), list(noise_levels)], names=['d', 'p'])
    df = DataFrame(columns=mi)
    for d in ds:
        code = code_class(
            d,
            noise='circuit-level',
            parametrization=parametrization,
            demolition=demolition,
            monolingual=monolingual,
            _merge_redundant_edges=_merge_redundant_edges,
        )
        decoder = decoder_class(
            code_class(d, noise='phenomenological'),
            **kwargs_for_decoder_class,
        )
        dc = {noise_level: code.SCHEME.run(
            decoder, noise_level, n) for noise_level, n in zip(noise_levels, ns)}
        df[d] = DataFrame(dc, index=('m', 'n'))
    return df

def subset_sample(
        ds: Sequence[int],
        noise_level: float,
        n: int,
        code_class: Type[Code],
        decoder_class: Type[Decoder],
        noise: NoiseModel,
        parametrization: Parametrization = 'balanced',
        tol: float = 5e-1,
):
    """Make threshold data for any decoder using subset sampling.
    
    
    :return df: A DataFrame indexed by (distance, error weight), with columns:
        1. 'subset prob' probability of error of that weight, given distance.
        2. 'survival prob' complement of cumulative sum of 'subset prob'.
        3. 'm' failure count.
        4. 'n' shot count.
    """
    ls = []
    for d in ds:
        code = code_class(
            d,
            noise=noise,
            parametrization=parametrization,
        )
        decoder = decoder_class(code)
        ls.append(decoder.subset_sample(noise_level, n, tol=tol))
    df = pd.concat(ls, keys=ds, names=['d'])
    return df


"""Plot failure probability to deduce accuracy thresholds.

Available functions:

* monte_carlo
* subset_sample
"""

import itertools

from matplotlib import pyplot as plt
from matplotlib.container import ErrorbarContainer
import numpy as np
from pandas import DataFrame
from statsmodels.stats.proportion import proportion_confint

from localuf.constants import STANDARD_ERROR_ALPHA


def monte_carlo(
        data: DataFrame,
        title: str = '',
        xlabel: None | str = None,
        ylabel: None | str = None,
        legend: None | bool = None,
        alpha: float = STANDARD_ERROR_ALPHA,
        method: str = 'wilson',
        base_color: None | tuple[float, float, float] | str = None,
        capsize: float = 2,
        **kwargs_for_errorbar,
):
    """Plot threshold data in ``data``.
    
    
    :param data: a DataFrame where each
        column is a (distance, probability);
        rows m, n indicate number of logical errors and samples, respectively.
    :param title: plot title.
    :param xlabel: x-axis label.
    :param ylabel: y-axis label.
    :param legend: whether to show legend.
    :param alpha: significance level of confidence intervals.
    :param method: method to compute confidence intervals.
        For details on confidence intervals,
        see https://www.statsmodels.org/dev/generated/statsmodels.stats.proportion.proportion_confint.html.
    :param base_color: a single color for all errorbars and their connecting lines.
        Increasing distance is then shown by increasing opacity.
        If ``None``, each distance is shown by a different, fully opaque color.
    :param capsize: length of error bar caps in points.
    :param kwargs_for_errorbar: passed to ``pyplot.errorbar``.
    
    
    :return plotted: Transposed ``data`` with additional columns ``f, lo, hi`` storing
        respectively the mean, lower and upper confidence bounds of failure probability.
    :return containers: A dictionary where each key is a distance;
        value, the ``ErrorbarContainer`` for that distance.
    """
    containers: dict[int, ErrorbarContainer] = {}
    if base_color is None:
        colors = itertools.repeat(None)
        if legend is None:
            legend = True
    else:
        if legend is None:
            legend = False
        d_count = len(data.columns.unique('d'))
        colors = [(base_color, k/d_count) for k in range(1, d_count+1)]
    plotted = data.T
    plotted['f'] = plotted.m / plotted.n
    plotted['lo'], plotted['hi'] = proportion_confint(
        plotted.m,
        plotted.n,
        alpha=alpha,
        method=method,
    )
    # next line is a fix for when m = 0 then sometimes lo > 0
    plotted['lo'] = np.min(plotted[['lo', 'f']], axis=1)
    for (d, df), color in zip(plotted.groupby(level='d'), colors):
        df = df.droplevel('d')
        container = plt.errorbar(
            x=df.index,
            y=df.f,
            yerr=(df.f-df.lo, df.hi-df.f),
            capsize=capsize,
            label=str(d),
            color=color,
            **kwargs_for_errorbar,
        )
        containers[d] = container # type: ignore
    if legend:
        plt.legend(title=r'$d =\dots$')
    if title:
        plt.title(title)
    if xlabel is None:
        xlabel = 'noise level'
    plt.xlabel(xlabel)
    if ylabel is None:
        ylabel = 'logical error probability'
    plt.ylabel(ylabel)
    return plotted, containers


def subset_sample(
        data: DataFrame,
        legend: bool = True,
        alpha: float = 0.3,
        title: str = '',
):
    """Plot failure probability from output of ``get_failure_data_from_subset_sample``.
    
    
    :param data: output of ``get_failure_data_from_subset_sample``.
    :param legend: whether to show legend.
    :param alpha: transparency of confidence region.
    :param title: plot title.
    """
    for d, df in data.groupby(level='d'):
        df = df.droplevel('d')
        plt.plot(df.index, df.f, label=str(d))
        plt.fill_between(
            x=df.index,
            y1=df.lo,
            y2=df.hi,
            alpha=alpha,
        )
    plt.loglog()
    if legend:
        plt.legend(title=r'$d =\dots$')
    plt.xlabel('noise level')
    plt.ylabel('logical error probability')
    if title:
        plt.title(title)


from localuf.type_aliases import Edge, Node


class Pairs:
    """A set of node pairs defining free anyon strings.
    
    In form of a dictionary where there are two entries, u:v and v:u, per pair uv.
    """

    def __repr__(self) -> str:
        return str(self.as_set)

    @property
    def as_set(self):
        """Return the set of pairs."""
        set_: set[Edge] = set()
        visited: set[Node] = set()
        for u, v in self._dc.items():
            if u not in visited:
                set_.add((u, v))
                visited.add(v)
        return set_

    def __init__(self) -> None:
        self._dc: dict[Node, Node] = {}

    def reset(self):
        """Factory reset."""
        self._dc.clear()

    def __contains__(self, u: Node):
        """Return whether u is in the set."""
        return u in self._dc

    def __getitem__(self, u: Node):
        """Return v if uv is in the set."""
        return self._dc[u]

    def add(self, u: Node, v: Node):
        """Add pair uv."""
        self._dc[u] = v
        self._dc[v] = u

    def remove(self, u: Node):
        """Remove pair containing u."""
        v = self._dc.pop(u)
        del self._dc[v]

    def load(self, e: Edge):
        """Load in edge ``e``."""
        u, v = e
        if u in self:
            w = self[u]
            self.remove(u)
            if v in self:
                x = self[v]
                self.remove(v)
                self.add(w, x)
            elif v != w:
                self.add(v, w)
        elif v in self:
            x = self[v]
            self.remove(v)
            self.add(u, x)
        else:
            self.add(u, v)


class LogicalCounter:
    """Counts logical error strings in, and updates, ``Pairs`` instance."""

    def __init__(
            self,
            d: int,
            commit_height: int,
            long_axis: int,
            time_axis: int,
    ) -> None:
        self._D = d
        self._COMMIT_HEIGHT = commit_height
        self._LONG_AXIS = long_axis
        self._TIME_AXIS = time_axis

    def _lower_node(self, v: Node) -> Node:
        """Move ``v`` down by commit height."""
        new_v = list(v)
        new_v[self._TIME_AXIS] -= self._COMMIT_HEIGHT
        return tuple(new_v)

    def count(self, pairs: Pairs):
        """Count logical error strings in, and update ``pairs``.
        
        
        ``pairs`` a set of node pairs defining free anyon strings.
        
        
        :returns:
            * ``error_count`` the number of logical error strings completed in ``pairs``.
            * ``new_pairs`` the error strings in ``pairs`` ending at the future boundary of the commit region, lowered by commit height.
        """
        error_count: int = 0
        new_pairs = Pairs()
        for u, v in pairs.as_set:
            pair_separation = abs(u[self._LONG_AXIS] - v[self._LONG_AXIS])
            if pair_separation == self._D:
                error_count += 1
            elif not (pair_separation == 0 and u[self._LONG_AXIS] in {-1, self._D-1}):
                new_pairs.load((
                    self._lower_node(u),
                    self._lower_node(v),
                ))
        return error_count, new_pairs


import itertools

import pytest

from localuf import Repetition
from localuf._schemes import Forward

@pytest.fixture(name="rp", params=itertools.product(
        range(3, 11, 2),
        ('phenomenological',),
        ('batch', 'forward', 'frugal'),
), ids=lambda x: f"d{x[0]} {x[2]}")
def _rp(request):
    d, noise, scheme = request.param
    return Repetition(d, noise, scheme=scheme)

@pytest.fixture(
        name="rpCC",
        params=range(3, 11, 2),
        ids=lambda x: f"d{x}",
)
def _rpCC(request):
    return Repetition(request.param, 'code capacity')


class TestPhenomenologicalEdges:

    @pytest.mark.parametrize("buffer_height", range(1, 4), ids=lambda x: f"buffer_height {x}")
    def test_t_start_consistency(self, forward_rp: Forward, buffer_height):
        code = forward_rp._CODE
        commit_height = forward_rp._COMMIT_HEIGHT
        n_commit_edges, commit_edges = code._phenomenological_edges(
            commit_height,
            True,
        )
        n_fresh_edges, fresh_edges = code._phenomenological_edges(
            commit_height,
            True,
            t_start=buffer_height,
        )
        assert n_commit_edges == len(commit_edges)
        assert n_fresh_edges == n_commit_edges
        assert fresh_edges == tuple((
            (*u[:-1], buffer_height+u[-1]),
            (*v[:-1], buffer_height+v[-1]),
        ) for u, v in commit_edges)
        nodes = set().union(*commit_edges)
        assert nodes == set(itertools.product(range(-1, code.D), range(commit_height))) | {
            (j, commit_height) for j in range(code.D-1)
        }

    def test_merge_equivalent_boundary_edges(self, forward_rp: Forward):
        code = forward_rp._CODE
        h = forward_rp.WINDOW_HEIGHT
        edge_count, edges = code._phenomenological_edges(
            h,
            False,
            merge_equivalent_boundary_nodes=True,
        )
        assert edge_count == len(edges)
        assert edge_count == code.D * h + (code.D-1) * (h-1)
        nodes = set().union(*edges)
        assert nodes == set(itertools.product(range(code.D-1), range(h))) | {
            (-1, 0), (code.D-1, 0),
        }


def test_N_EDGES_attribute(rp: Repetition):
    assert type(rp.N_EDGES) is int
    assert rp.N_EDGES == len(rp.EDGES)

def test_EDGES_attribute(rp: Repetition):
    assert type(rp.EDGES) is tuple

def test_NODES_code_capacity(rpCC: Repetition):
    assert type(rpCC.NODES) is tuple
    assert len(rpCC.NODES) == rpCC.D+1

def test_NODES_attribute(rp: Repetition):
    assert type(rp.NODES) is tuple
    match str(rp.SCHEME):
        case 'batch':
            assert len(rp.NODES) == rp.D * (rp.D+1)
        case 'forward' | 'frugal':
            assert len(rp.NODES) == \
                (rp.D+1) * rp.SCHEME.WINDOW_HEIGHT + rp.D-1

def test_LONG_AXIS_attribute(rp: Repetition):
    assert rp.LONG_AXIS == 0

def test_DIMENSION_attribute(rp: Repetition):
    assert rp.DIMENSION == len(rp.NODES[0])

def test_get_pos():
    rp = Repetition(3, 'code capacity')
    pos = rp.get_pos()
    assert type(pos) is dict


import pytest

from localuf.type_aliases import Node
from localuf.decoders.luf import Macar

# edit be independent of luf!
@pytest.mark.parametrize("decoder, syndrome", [
    ("macar5F", "syndrome5F"),
    ("macar5T", "syndrome5T"),
])
def test_load(decoder: Macar, syndrome: set[Node], request):
    decoder = request.getfixturevalue(decoder)
    syndrome = request.getfixturevalue(syndrome)
    decoder.NODES.load(syndrome)
    for v, node in decoder.NODES.dc.items():
        if v in syndrome:
            assert node.defect
            assert node.anyon
            assert node.active
        else:
            assert not node.defect
            assert not node.anyon
            assert not node.active


from collections.abc import Callable
from typing import Iterable, Literal, Type

from pandas import DataFrame

from localuf._schemes import Frugal
from localuf.type_aliases import StreamingNoiseModel
from localuf._base_classes import Code
from localuf.sim._height_calculator import get_heights
from localuf.decoders.snowflake import Snowflake


def frugal(
        ds: Iterable[int],
        noise_levels: Iterable[float],
        n: int,
        code_class: Type[Code],
        noise: StreamingNoiseModel,
        time_only: Literal['all', 'merging', 'unrooting'] = 'merging',
        get_commit_height: Callable[[int], int] | None = None,
        get_buffer_height: Callable[[int], int] | None = None,
        **kwargs_for_Snowflake,
):
    """Make latency data Snowflake.
    
    
    :param ds: an iterable of surface code distances.
    :param noise_levels: an iterable of noise levels.
    :param n: sample count.
    :param code_class: the class of the code.
    :param noise: the noise model.
    :param time_only: whether runtime includes a timestep
        for each drop, each grow, and each merging step ('all');
        each merging step only ('merging');
        or each unrooting step only ('unrooting').
    :param get_commit_height: a function with input ``d`` that outputs commit height.
        If ``None``, commit height is ``1``.
    :param get_buffer_height: a function with input ``d`` that outputs buffer height.
        If ``None``, buffer height is ``2*(d//2)``.
    :param kwargs_for_Snowflake: passed to Snowflake
        e.g. ``merger`` decides whether Snowflake's nodes
        flood before syncing (fast) or vice versa (slow) in a merging step.
    
    Output: ``data`` a DataFrame where each
    column a (distance, probability);
    row, a latency sample.
    """
    scheme = 'frugal'
    dc: dict[tuple[int, float], list[int]] = {}
    for d in ds:
        _, commit_height, buffer_height = get_heights(
            d,
            n,
            scheme,
            get_commit_height=get_commit_height,
            get_buffer_height=get_buffer_height,
        )
        code = code_class(
            d,
            noise,
            scheme=scheme,
            window_height=None,
            commit_height=commit_height,
            buffer_height=buffer_height,
        )
        frugal: Frugal = code.SCHEME # type: ignore
        decoder = Snowflake(code, **kwargs_for_Snowflake)
        for noise_level in noise_levels:
            dc[d, noise_level] = [frugal.sample_latency(
                decoder, noise_level, time_only=time_only) for _ in range(n)]
    data = DataFrame(dc).sort_index(axis=1)
    data.columns.set_names(['d', 'p'], inplace=True)
    return data


from unittest import mock

import pytest

from localuf.constants import Growth
from localuf.decoders import Snowflake
from localuf.decoders.snowflake import _Edge


@pytest.fixture
def sfe3a(snowflake3: Snowflake):
    """`_Edge` instance with `EdgeContact`."""
    return snowflake3.EDGES[(0, 1), (1, 1)]


@pytest.mark.parametrize("prop", [
    "DROPEE",
])
def test_property_attributes(test_property, sfe3a: _Edge, prop):
    test_property(sfe3a.CONTACT, prop)


def test_init(sfe3a: _Edge):
    with mock.patch("localuf.decoders.snowflake._Contact.__init__") as mock_init:
        sfe3a.CONTACT.__init__(sfe3a)
        mock_init.assert_called_once_with(sfe3a)


def test_drop(snowflake3: Snowflake):

    edge_1 = snowflake3.EDGES[(0, 1), (1, 1)]
    edge_0 = snowflake3.EDGES[(0, 0), (1, 0)]

    edge_1.growth = Growth.FULL
    edge_1.correction = True

    edge_1.CONTACT.drop()
    
    assert edge_0.next_growth is Growth.FULL
    assert edge_0.next_correction is True


from unittest import mock

import pytest

from localuf._base_classes import Code
from localuf.codes import Repetition, Surface
from localuf.constants import Growth
from localuf.decoders.snowflake.constants import RESET
from localuf.type_aliases import Node
from localuf.decoders import Snowflake
from localuf.decoders.snowflake import _Node, NothingFriendship, NodeFriendship, TopSheetFriendship
from localuf.decoders._base_uf import direction

@pytest.mark.parametrize("prop", [
    "SNOWFLAKE",
    "INDEX",
    "ID",
    "FRIENDSHIP",
    "NEIGHBORS",
    "UNROOTER",
])
def test_property_attributes(test_property, sfn3: _Node, prop):
    test_property(sfn3, prop)


def test_init(sfn3: _Node):
    with mock.patch("localuf.decoders.snowflake._Node.reset") as mock_reset:
        sfn3.__init__(sfn3.SNOWFLAKE, sfn3.INDEX)
        mock_reset.assert_called_once_with()


def test_SNOWFLAKE_property(sfn3: _Node):
    assert type(sfn3.SNOWFLAKE) is Snowflake
    assert sfn3.SNOWFLAKE.NODES[sfn3.INDEX] == sfn3


def test_INDEX_property(sfn3: _Node, fixture_test_INDEX_property):
    fixture_test_INDEX_property(sfn3)


def test_ID_property(sfn3: _Node):
    assert type(sfn3.ID) is int


def test_FRIENDSHIP_property(snowflake: Snowflake):
    for v, node in snowflake.NODES.items():
        if v[snowflake.CODE.TIME_AXIS] == 0:
            assert type(node.FRIENDSHIP) is NothingFriendship
        elif v[snowflake.CODE.TIME_AXIS] == snowflake.CODE.SCHEME.WINDOW_HEIGHT-1:
            assert type(node.FRIENDSHIP) is TopSheetFriendship
        else:
            assert type(node.FRIENDSHIP) is NodeFriendship


class TestNEIGHBORSProperty:

    def _general_checks(self, code: Code, decoder: Snowflake):
        for v, node in decoder.NODES.items():
            assert type(node.NEIGHBORS) is dict
            is_top_detector = (
                v[code.TIME_AXIS]==code.SCHEME.WINDOW_HEIGHT-1
            ) and not node._IS_BOUNDARY
            if isinstance(code, Repetition):
                assert len(node.NEIGHBORS) + is_top_detector == len(code.INCIDENT_EDGES[v])
            else:
                if is_top_detector:
                    assert len(node.NEIGHBORS) < len(code.INCIDENT_EDGES[v])
                else:
                    assert len(node.NEIGHBORS) == len(code.INCIDENT_EDGES[v])


    def test_default_order(self, snowflake3: Snowflake, sfn3: _Node):
        self._general_checks(snowflake3.CODE, snowflake3)
        assert sfn3.NEIGHBORS == {
            'W': (((-1, 0), (0, 0)), 0),
            'E': (((0, 0), (1, 0)), 1),
            'U': (((0, 0), (0, 1)), 1),
        }


    def test_repetition_custom_order(self, frugal_rep_3: Repetition):
        decoder = Snowflake(
            code=frugal_rep_3,
            _neighbor_order=('U', 'W', 'E', 'D')
        )
        self._general_checks(frugal_rep_3, decoder)
        assert tuple(decoder.NODES[0, 0].NEIGHBORS.items()) == (
            ('U', (((0, 0), (0, 1)), 1)),
            ('W', (((-1, 0), (0, 0)), 0)),
            ('E', (((0, 0), (1, 0)), 1)),
        )

    def test_surface_custom_order(self, surface3_CL_frugal: Surface):
        decoder = Snowflake(
            surface3_CL_frugal,
            _neighbor_order=(
                'U',
                'NU',
                'EU',
                'SEU',
                'N',
                'W',
                'E',
                'S',
                'NWD',
                'WD',
                'SD',
                'D',
            )
        )
        self._general_checks(surface3_CL_frugal, decoder)
        assert tuple(decoder.NODES[0, 0, 0].NEIGHBORS.items()) == (
            ('U', (((0, 0, 0), (0, 0, 1)), 1)),
            ('EU', (((0, 0, 0), (0, 1, 1)), 1)),
            ('SEU', (((0, 0, 0), (1, 1, 1)), 1)),
            ('W', (((0, -1, 0), (0, 0, 0)), 0)),
            ('E', (((0, 0, 0), (0, 1, 0)), 1)),
            ('S', (((0, 0, 0), (1, 0, 0)), 1)),
        )


def test_busy_attribute(sfn3: _Node):
    assert sfn3.busy is False


def test_reset(sfn3: _Node):

    sfn3.defect = True
    sfn3.active = True
    sfn3.cid = -1
    sfn3.pointer = 'U'
    sfn3.busy = True

    sfn3.next_defect = True
    sfn3.next_active = True
    sfn3.next_cid = -1

    sfn3.access = {'U': sfn3.SNOWFLAKE.NODES[(0, 1)]}

    sfn3.reset()

    assert sfn3.defect is False
    assert sfn3.active is False
    assert sfn3.cid == sfn3.ID
    assert sfn3.pointer == 'C'
    assert sfn3.busy is False

    assert sfn3.next_defect is False
    assert sfn3.next_active is False
    assert sfn3.next_cid == sfn3.ID

    assert sfn3.access == {}


def test_update_after_drop(sfn3: _Node):
    sfn3.next_defect = True
    sfn3.next_active = True
    sfn3.next_cid = RESET
    sfn3.next_pointer = 'U'

    sfn3.update_after_drop()

    assert sfn3.defect
    assert sfn3.active
    assert sfn3.cid == RESET
    assert sfn3.pointer == 'U'
    
    assert sfn3.next_defect
    assert sfn3.next_active
    assert sfn3.next_cid == RESET
    assert sfn3.next_pointer == 'U'


@pytest.mark.parametrize("active", (False, True))
def test_grow(sfn3: _Node, active):
    """Test `grow`."""
    sfn3.active = active
    with (
        mock.patch("localuf.decoders.snowflake._Node._grow") as mock_grow,
        mock.patch("localuf.decoders.snowflake.NothingFriendship.find_broken_pointers") as mock_find,
    ):
        sfn3.grow()
        assert sfn3.busy is False
        if active:
            mock_grow.assert_called_once_with()
        else:
            mock_grow.assert_not_called()
        mock_find.assert_called_once_with()


@pytest.mark.parametrize("active", (False, True))
@pytest.mark.parametrize("whole", (False, True))
def test_grow_whole(sfn3: _Node, active, whole):
    sfn3.active = active
    sfn3.whole = whole
    with (
        mock.patch("localuf.decoders.snowflake._Node._grow") as mock_grow,
        mock.patch("localuf.decoders.snowflake.NothingFriendship.find_broken_pointers") as mock_find,
    ):
        sfn3.grow_whole()
        assert sfn3.busy is False
        if active and whole:
            mock_grow.assert_called_once_with()
            assert sfn3.grown
            assert sfn3.next_grown
        else:
            mock_grow.assert_not_called()
            assert not sfn3.grown
            assert not sfn3.next_grown
        mock_find.assert_called_once_with()


@pytest.mark.parametrize("active", (False, True))
@pytest.mark.parametrize("whole", (False, True))
@pytest.mark.parametrize("grown", (False, True))
def test_grow_half(sfn3: _Node, active, whole, grown):
    sfn3.active = active
    sfn3.whole = whole
    sfn3.grown = grown
    sfn3.unrooted = True
    sfn3.next_unrooted = True
    with (
        mock.patch("localuf.decoders.snowflake._Node._grow") as mock_grow,
        mock.patch("localuf.decoders.snowflake.NothingFriendship.find_broken_pointers") as mock_find,
    ):
        sfn3.grow_half()
        assert not sfn3.unrooted
        assert not sfn3.next_unrooted
        assert sfn3.busy is False
        if active and not whole and not grown:
            mock_grow.assert_called_once_with()
        else:
            mock_grow.assert_not_called()
        assert sfn3.grown is grown
        assert sfn3.next_grown is False
        mock_find.assert_not_called()


def test_grow_subroutine(sfn3: _Node):
    """Test `_grow`."""
    for growth in [Growth.HALF] + 2*[Growth.FULL]:
        sfn3._grow()
        assert sfn3.busy is False
        for e, edge in sfn3.SNOWFLAKE.EDGES.items():
            if sfn3.INDEX in e:
                assert edge.growth is growth
            else:
                assert edge.growth is Growth.UNGROWN


def test_update_access(snowflake3: Snowflake):
    w, c, e = (-1, 0), (0, 0), (1, 0)
    center = snowflake3.NODES[c]
    center.update_access()
    assert center.access == {}
    snowflake3.EDGES[w, c].growth = Growth.FULL
    center.update_access()
    assert center.access == {'W': snowflake3.NODES[w]}
    snowflake3.EDGES[c, e].growth = Growth.FULL
    center.update_access()
    assert center.access == {
        'W': snowflake3.NODES[w],
        'E': snowflake3.NODES[e]
    }
    snowflake3.EDGES[w, c].growth = Growth.UNGROWN
    center.update_access()
    assert center.access == {'E': snowflake3.NODES[e]}


@pytest.mark.parametrize("whole", (False, True))
def test_merging(sfn3: _Node, whole):
    sfn3.busy = True
    with (
        mock.patch("localuf.decoders.snowflake._Node.syncing") as mock_syncing,
        mock.patch("localuf.decoders.snowflake._Node.flooding") as mock_flooding,
    ):
        sfn3.merging(whole)
        assert sfn3.busy is False
        mock_syncing.assert_called_once_with()
        mock_flooding.assert_called_once_with(whole)


def test_syncing(syncing_flooding_objects: tuple[
    Snowflake, tuple[Node, Node, Node], tuple[_Node, _Node, _Node]
]):
    snowflake3, (w, c, e), (west, center, east) = syncing_flooding_objects

    # TEST PUSH DEFECT
    center.pointer = 'E'
    center.defect = True
    center.next_defect = True
    center.access = {'E': east}
    center.syncing()
    assert center.busy is True
    assert east.next_defect is True
    assert center.defect is True
    assert center.next_defect is False
    assert snowflake3.EDGES[c, e].correction is True
    snowflake3.reset()

    # TEST BOUNDARY PUSHES NOT ITS DEFECT
    west.pointer = 'E'
    west.defect = True
    west.next_defect = True
    west.access = {'E': center}
    west.syncing()
    assert west.busy is False
    assert center.next_defect is False
    assert west.defect is True
    assert west.next_defect is True
    assert snowflake3.EDGES[w, c].correction is False
    snowflake3.reset()

    # TEST UPDATE ACTIVE
    # boundary root never active
    for defect in (False, True):
        west.defect = defect
        west.syncing()
        assert west.next_active is False
        assert west.busy is False
        west.reset()
    # nonboundary root without defect
    center.syncing()
    assert center.next_active is False
    assert center.busy is False
    center.reset()
    # nonboundary root with defect
    center.defect = True
    center.syncing()
    assert center.next_active is True
    assert center.busy is True
    center.reset()
    # else
    center.pointer = 'E'
    center.access = {'E': east}
    east.active = True
    center.syncing()
    assert center.next_active is True
    assert center.busy is True


@pytest.mark.parametrize("whole", (False, True))
def test_flooding(sfn3: _Node, whole):
    with (
        mock.patch("localuf.decoders.snowflake.main._FullUnrooter.flooding_whole") as mock_whole,
        mock.patch("localuf.decoders.snowflake.main._FullUnrooter.flooding_half") as mock_half,
    ):
        sfn3.flooding(whole)
        if whole:
            mock_whole.assert_called_once_with()
            mock_half.assert_not_called()
        else:
            mock_whole.assert_not_called()
            mock_half.assert_called_once_with()


def test_update_after_merging(sfn3: _Node):
    sfn3.next_defect = True
    sfn3.next_cid = -1
    sfn3.next_defect = True
    sfn3.next_active = True
    sfn3.update_after_merging()
    assert sfn3.cid == -1
    assert sfn3.defect
    assert sfn3.active
    assert sfn3.next_defect
    assert sfn3.next_active


from localuf.type_aliases import Node, Edge
from localuf._pairs import Pairs

FourNodes = tuple[Node, Node, Node, Node]


def test_load_annihilates_anyons(pairs: Pairs, uvwx: FourNodes):
    u, v, _, _ = uvwx
    pairs._dc = {u: v, v: u}
    pairs.load((u, v))
    assert pairs._dc == {}


def test_load_joins_paths(pairs: Pairs, uvwx: FourNodes):
    u, v, w, x = uvwx
    pairs._dc = {u: w, w: u, v: x, x: v}
    pairs.load((u, v))
    assert pairs._dc == {w: x, x: w}


def test_load_extends_path(pairs: Pairs, uvwx: FourNodes):
    u, v, w, _ = uvwx
    pairs._dc = {u: w, w: u}
    pairs.load((u, v))
    assert pairs._dc == {v: w, w: v}
    # test symmetry
    pairs.load((u, v))
    assert pairs._dc == {u: w, w: u}


def test_load_adds_path(pairs: Pairs, uvwx: FourNodes):
    u, v, w, x = uvwx
    pairs.load((u, v))
    assert pairs._dc == {u: v, v: u}
    pairs.load((w, x))
    assert pairs._dc == {u: v, v: u, w: x, x: w}


def test_load_ignores_cycle(pairs: Pairs, uvwx: FourNodes):
    u, v, w, x = uvwx
    a_cycle: set[Edge] = {
        (u, v),
        (v, x),
        (w, x),
        (u, w),
    }
    for e in a_cycle:
        pairs.load(e)
    assert pairs._dc == {}


from math import prod

import numpy as np
import pytest
from scipy.stats import binom

from localuf.noise.forcers import _BaseForceByPair
from localuf.type_aliases import FourInts, FourFloats

@pytest.fixture
def base_force_by_pair():
    toy_m: FourInts = tuple(range(4)) # type: ignore
    bfbp = _BaseForceByPair(edges={
        toy_m: [((0, -1, 0), (0, 0, 0)), ((0, 0, 0), (0, 0, 1))],
    })
    bfbp.PAIR_POPULATIONS = (
        bfbp._EDGES[toy_m],
        bfbp._EDGES[toy_m],
    )
    return bfbp


def test_subset_probability(base_force_by_pair: _BaseForceByPair):
    weights = ((0, 0), (1, 0), (0, 1), (1, 1))
    pi: FourFloats = (0.1, 0.2, 0.3, 0.4)
    subset_prob: np.ndarray = base_force_by_pair.subset_probability(
        weights=weights,
        pi=pi,
    ) # type: ignore
    assert (subset_prob == np.array([
        prod(binom.pmf(w, 2, p) for w, p in zip(weight, pi))
        for weight in weights
    ])).all()


from typing import Type
from unittest.mock import patch

from pandas import DataFrame, MultiIndex
import pytest

from localuf import Repetition, Surface
from localuf.decoders.snowflake import Snowflake
from localuf.sim import runtime
from localuf._base_classes import Code
from localuf._schemes import Frugal


@pytest.mark.parametrize('code_class', (Repetition, Surface))
def test_frugal(code_class: Type[Code]):
    ds = [1, 2]
    noise_levels = [0, 1]
    n = 2
    call_count = len(ds) * len(noise_levels)

    def make_step_counts(
            decoder: Snowflake,
            noise_level: float,
            n: int,
            **_,
    ):
        frugal: Frugal = decoder.CODE.SCHEME # type: ignore
        frugal.step_counts = (n-1)*[int(noise_level)]

    with patch(
        'localuf._schemes.Frugal.run',
        side_effect=make_step_counts,
    ) as mock_run:
        data = runtime.frugal(
            ds=ds,
            noise_levels=noise_levels,
            n=n,
            code_class=code_class,
            noise='phenomenological',
        )
        assert mock_run.call_count == call_count
    assert type(data) is DataFrame
    assert data.shape == (n-1, len(ds) * len(noise_levels))
    assert data.columns.names == ['d', 'p']
    assert (data.columns == MultiIndex.from_product([ds, noise_levels])).all()


from unittest import mock

import pytest

from localuf.decoders.luf import Macar, MacarNode
from localuf.decoders.luf.constants import Stage

@pytest.fixture
def vn(macar: Macar):
    v = (0, 0) if macar.NODES.LUF.CODE.DIMENSION==2 else (0, 0, 0)
    return MacarNode(macar.NODES, v)

def test_advance(vn: MacarNode):
    # from https://stackoverflow.com/a/63690318/20887677
    with mock.patch("localuf.decoders.luf.MacarNode.growing") as growing_mock:
        vn.advance()
        growing_mock.assert_called_once_with()
    vn.NODES.LUF.CONTROLLER.stage = Stage.MERGING
    with mock.patch("localuf.decoders.luf.MacarNode.merging") as merging_mock:
        vn.advance()
        merging_mock.assert_called_once_with()
    vn.NODES.LUF.CONTROLLER.stage = Stage.PRESYNCING
    with mock.patch("localuf.decoders.luf.MacarNode.presyncing") as presyncing_mock:
        vn.advance()
        presyncing_mock.assert_called_once_with()
    vn.NODES.LUF.CONTROLLER.stage = Stage.SYNCING
    with mock.patch("localuf.decoders.luf.MacarNode.syncing") as syncing_mock:
        vn.advance()
        syncing_mock.assert_called_once_with()